(define-module (sx-ext e-lib))
(use-modules  (ice-9 regex)
              (sx undo) (sx property) (sx basic) 
              (sx selection-basic) (sx view) (sx render)
              (sx gui-gl) (sx gui-entry) (sx gui-optmenu)
)

(export command-e-lib-read command-e-lib-write)


; add the commands to the sx system
(entry-add-command-list '("e-lib-read" "e-lib-write"))


(define (command-e-lib-read)
  (selection-setup selection-string
                   operation-e-lib-read
                   "Enter filename (with path, no extension):"
                   *OPTION_EMPTY* 'view #f #f)
)

(define (command-e-lib-write)
  (selection-setup selection-string
                   operation-e-lib-write
                   "Enter filename (with path, no extension):"
                   *OPTION_EMPTY* 'view #f #f)
)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; private functions

; writes the header file of the e-lib format 
(define (e-lib-write-header fp)
  (display "TITLE              \"Generated by SX\"\n" fp)
  (display "NMESHPOINTS        "fp)
  (display (length *POINTS*) fp)(newline fp)
  (display "NNODES             "fp)
  (display (length *POINTS*) fp)(newline fp)
  (if (> (length *LINKS*) 0)
    (begin
      (display "NELEMENTS_LINK1    "fp)
      (display (length *LINKS*) fp)(newline fp)
    )
  )
  (if (> (length *TRIANGS*) 0)
    (begin
      (display "NELEMENTS_TRIANG1  "fp)
      (display (length *TRIANGS*) fp)(newline fp)
    )
  )
  (if (> (length *QUADS*) 0)
    (begin
      (display "NELEMENTS_QUAD1    "fp)
      (display (length *QUADS*) fp)(newline fp)
    )
  )
  (newline fp)(newline fp)
)


; writes the meshpoints
(define (e-lib-write-meshpoints fp)
  (display "MESHPOINT_COORDINATES\n" fp)
  (let*
    ( (n 1) (co #f)
      (point-map (map (lambda (x)
                        (display n fp)(display "  " fp)
                        (set! n (1+ n))
                        (set! co (object-property-get 'p (car x) 'coords))
                        (display (car   co) fp)(display "  " fp)
                        (display (cadr  co) fp)(display "  " fp)
                        (display (caddr co) fp)(newline fp)
                        (car x)
                      )
                      *POINTS*
                 )
      )
    )
    (newline fp)(newline fp)
    point-map
  )
)


; writes the elements, link, triangles, quads
(define (e-lib-write-elements point-map fp)
  (set! point-map (reverse point-map))
  ; links
  (if (> (length *LINKS*) 0)
    (let* ( (n 1) (nodes #f) )
      (display "NODES_LINK1\n"fp)
      (for-each (lambda (x)
          (display n fp)(display "  " fp)
          (set! n (1+ n))
          (set! nodes (object-property-get 'l (car x) 'nodes))
          (display (length (member (car nodes) point-map)) fp)
          (display "  " fp)
          (display (length (member (cadr nodes) point-map)) fp)
          (newline fp)
        )
        *LINKS*
      )
      (newline fp)(newline fp)
    )
  )
  ; triangles
  (if (> (length *TRIANGS*) 0)
    (let* ( (n 1) (nodes #f) )
      (display "NODES_TRIANG1\n"fp)
      (for-each (lambda (x)
          (display n fp)(display "  " fp)
          (set! n (1+ n))
          (set! nodes (object-property-get 't (car x) 'nodes))
          (display (length (member (car nodes) point-map)) fp)
          (display "  " fp)
          (display (length (member (cadr nodes) point-map)) fp)
          (display "  " fp)
          (display (length (member (caddr nodes) point-map)) fp)
          (newline fp)
        )
        *TRIANGS*
      )
      (newline fp)(newline fp)
    )
  )
  ; quads
  (if (> (length *QUADS*) 0)
    (let* ( (n 1) (nodes #f) )
      (display "NODES_QUAD1\n"fp)
      (for-each (lambda (x)
          (display n fp)(display "  " fp)
          (set! n (1+ n))
          (set! nodes (object-property-get 'q (car x) 'nodes))
          (display (length (member (car nodes) point-map)) fp)
          (display "  " fp)
          (display (length (member (cadr nodes) point-map)) fp)
          (display "  " fp)
          (display (length (member (caddr nodes) point-map)) fp)
          (display "  " fp)
          (display (length (member (cadddr nodes) point-map)) fp)
          (newline fp)
        )
        *QUADS*
      )
      (newline fp)(newline fp)
    )
  )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; reads in a non-empty line, comments are ignored
(define (e-lib-read-line fp)
  (let* ( (buffer (make-string 80))
          (curr-buf-len (string-length buffer))
          (increment 8) 
          (i 0)
          (c #f)
          (read-more #t)
        )
    (while read-more
      (set! c (read-char fp))
      (if (not (eof-object? c))
        (begin
          ; first non-space character
          (while (and (not (eof-object? c)) (char-whitespace? c))
            (set! c (read-char fp))
          )
          (cond
            ; single line comment
            ( (equal? c #\#)
              (while (not (or (eof-object? c) (equal? c #\newline)))
                (set! c (read-char fp))
              )
            )
            ; multiple line comment
            ( (and (equal? c #\/) (equal? (peek-char fp) #\*))
              ; skip until end of it
              (while (not (or (eof-object? c) 
                              (and (equal? c #\*)
                                   (equal? (peek-char fp) #\/))))
                (set! c (read-char fp))
              )
              ; skip until end of line
              (while (not (or (eof-object? c) (equal? c #\newline)))
                (set! c (read-char fp))
              )
            )
            ( else
              (set! read-more #f)
              (while (not (or (eof-object? c) (equal? c #\newline)))
                (if (>= i curr-buf-len)
                  (begin
                    (set! buffer (string-append buffer 
                                                (make-string increment)))
                    (set! curr-buf-len (string-length buffer))
                  )
                )
                (string-set! buffer i c)
                (set! i (1+ i))
                (set! c (read-char fp))
              )
            )
          )
        )
        ; eof file
        (set! read-more #f)
      )
    ) ; end of while
     
    (if (> i 0)
      (begin
;        (display "line: ")(display (substring buffer 0 i))(newline)
        (substring buffer 0 i)
      )
      'eof
    )
  ) ; end of let 
)


; reads in an optional or mandatory keyword
; optional keywords can be ignored
(define (e-lib-read-keyword fp rx-list)
  ; if keyword is optional 
  (if (cdar rx-list)
    (call-with-current-continuation
      (lambda (found-optional)
        (do ((line (e-lib-read-line fp)
                   (e-lib-read-line fp)))
            ((equal? line 'eof))
          (call-with-current-continuation
            (lambda (end-optional)
              (for-each 
                (lambda (l-rx-type)
                  (let*
                    ( (l-rx  (car l-rx-type))
                      (l-opt (cdr l-rx-type))
                      (l-m   (regexp-exec l-rx line))
                    )
                    (if l-m
                      (found-optional l-m)
                    )
                    (if (not l-opt)
                      (end-optional #f)
                    )
                  )
                )
                rx-list
              )
            )
          )
        )
        #f
      )
    )
    ; mandatory keyword
    (call-with-current-continuation
      (lambda (found-mandatory)
        (let* ( (l-m #f) )
          (do ((line (e-lib-read-line fp)
                     (e-lib-read-line fp)))
              ((equal? line 'eof))
            (set! l-m (regexp-exec (caar rx-list) line))
            (if l-m
              (begin
                (found-mandatory l-m)
              )
            )
          )
        )
      )
    )
  )
)


; reads in a series of numbers with an index
; proc determines whether it is number? or integer?
(define (e-lib-read-number-series proc i n fp)
  (call-with-current-continuation
    (lambda (return-n-series)
      (let*
        ( (l-line (e-lib-read-line fp))
          (res (catch #t 
                      (lambda ()
                        (call-with-input-string
                          (string-append "("
                                         l-line
                                         ")")
                          read)
                      )
                      (lambda args #f)))
        )
        (if (not res)
          (begin
            (display "invalid line: ")
            (display l-line)(newline)
            (return-n-series #f)
          )
        )
        (if (not (integer? (car res)))
          (begin
            (display "invalid index: ")
            (display l-line)(newline)
            (return-n-series #f)
          )
        )
        (if (not (= (1+ i) (car res)))
          (begin
            (display "index out of sequence: ")
            (display l-line)(newline)
            (return-n-series #f)
          )
        )
        (if (not (= (length (cdr res)) n))
          (begin
            (display "not enough number: ")
            (display l-line)(newline)
            (return-n-series #f)
          )
        )
        (for-each (lambda (x) 
            (if (not (proc x))
              (begin
                (display "invalid line: ")
                (display l-line)(newline)
                (return-n-series #f)
              )
            )
          )
          (cdr res)
        )
        (cdr res)
      )
    )
  )
)


(define (e-lib-read-meshpoints npoint fp)
  (call-with-current-continuation
    (lambda (return-meshpoint)
      (let ( (coord #f) )
        (do ((i 0 (1+ i))) ((= i npoint))
          (set! coord (e-lib-read-number-series number? i 3 fp))
          (if (not coord)
            (return-meshpoint #f)
          )
          ; ready to use the coordinates
          (point-add-basic #f coord #f #t)
        )
      )
      #t
    )
  )
)

(define (e-lib-read-links nlink fp)
  (call-with-current-continuation
    (lambda (return-link)
      (let ( (nodes #f) )
        (do ((i 0 (1+ i))) ((= i nlink))
          (set! nodes (e-lib-read-number-series integer? i 2 fp))
          (if (not nodes)
            (return-link #f)
          )
          ; ready to use the nodes
          (link-add-basic #f nodes #f #t)
        )
      )
      #t
    )
  )
)

(define (e-lib-read-triangs ntriang fp)
  (call-with-current-continuation
    (lambda (return-triang)
      (let ( (nodes #f) )
        (do ((i 0 (1+ i))) ((= i ntriang))
          (set! nodes (e-lib-read-number-series integer? i 3 fp))
          (if (not nodes)
            (return-triang #f)
          )
          ; ready to use the nodes
          (triang-add-basic #f nodes #f #t)
        )
      )
      #t
    )
  )
)

(define (e-lib-read-quads nquad fp)
  (call-with-current-continuation
    (lambda (return-quad)
      (let ( (nodes #f) )
        (do ((i 0 (1+ i))) ((= i nquad))
          (set! nodes (e-lib-read-number-series integer? i 4 fp))
          (if (not nodes)
            (return-quad #f)
          )
          ; ready to use the nodes
          (quad-add-basic #f nodes #f #t)
        )
      )
      #t
    )
  )
)

; reads the e-lib mesh, 
; basic geometry and topological information
(define (e-lib-read-mesh fp)
  (let* 
    ( (line #f)
      (rx-point  (make-regexp "NMESHPOINTS[ \t]+"))
      (rx-coord  (make-regexp "MESHPOINT_COORDINATES[ \t]*"))
      (rx-link   (make-regexp "NELEMENTS_LINK1[ \t]+"))
      (rx-triang (make-regexp "NELEMENTS_TRIANG1[ \t]+"))
      (rx-quad   (make-regexp "NELEMENTS_QUAD1[ \t]+"))
      (rx-l-node (make-regexp "NODES_LINK1[ \t]*"))
      (rx-t-node (make-regexp "NODES_TRIANG1[ \t]*"))
      (rx-q-node (make-regexp "NODES_QUAD1[ \t]*"))
      (rx-list 
        (list (cons rx-point #f)
              (cons rx-link #t)
              (cons rx-triang #t)
              (cons rx-quad #t)
              (cons rx-coord #f)
        )
      )
      (m #f)
      (npoint 0) (nlink 0) (ntriang 0) (nquad 0)
    )
    (call-with-current-continuation
      (lambda (return)
        (while (not (null? rx-list))
          (set! m (e-lib-read-keyword fp rx-list))
          (if m
            (let* 
              ( (str (match:string m)) 
              )
              (cond
                ; set number of meshpoints
                ( (regexp-exec rx-point str)
                  (set! npoint (string->number (substring str (match:end m))))
                  (if (not npoint)
                    (begin
                      (display "number of meshpoints is not a number")(newline)
                      (return #f)
                    )
                  )
                  (set! rx-list (cdr (member (cons rx-point #f) rx-list)))
                )
                ; set number of links
                ( (regexp-exec rx-link str)
                  (set! nlink (string->number (substring str (match:end m))))
                  (if (not nlink)
                    (begin
                      (display "number of links is not a number")(newline)
                      (return #f)
                    )
                  )
                  (set! rx-list (append (cdr (member (cons rx-link #t) 
                                                     rx-list)
                                        )
                                        (list (cons rx-l-node #f))))
                )
                ; set number of triangs
                ( (regexp-exec rx-triang str)
                  (set! ntriang (string->number (substring str (match:end m))))
                  (if (not ntriang)
                    (begin
                      (display "number of triangs is not a number")(newline)
                      (return #f)
                    )
                  )
                  (set! rx-list (append (cdr (member (cons rx-triang #t) 
                                                     rx-list)
                                        )
                                        (list (cons rx-t-node #f))))
                )
                ; set number of quads
                ( (regexp-exec rx-quad str)
                  (set! nquad (string->number (substring str (match:end m))))
                  (if (not nquad)
                    (begin
                      (display "number of quads is not a number")(newline)
                      (return #f)
                    )
                  )
                  (set! rx-list (append (cdr (member (cons rx-quad #t) 
                                                     rx-list)
                                        )
                                        (list (cons rx-q-node #f))))
                )
                ; read the coordinates
                ( (regexp-exec rx-coord str)
                  (if (not (e-lib-read-meshpoints npoint fp))
                    (return #f)
                  )
                  (set! rx-list (cdr (member (cons rx-coord #f) rx-list)))
                )
                ; read link elements
                ( (regexp-exec rx-l-node str)
                  (if (not (e-lib-read-links nlink fp))
                    (return #f)
                  )
                  (set! rx-list (cdr (member (cons rx-l-node #f) rx-list)))
                )
                ; read triang elements
                ( (regexp-exec rx-t-node str)
                  (if (not (e-lib-read-triangs ntriang fp))
                    (return #f)
                  )
                  (set! rx-list (cdr (member (cons rx-t-node #f) rx-list)))
                )
                ; read quad elements
                ( (regexp-exec rx-q-node str)
                  (if (not (e-lib-read-quads nquad fp))
                    (return #f)
                  )
                  (set! rx-list (cdr (member (cons rx-q-node #f) rx-list)))
                )
                ( else 
                  (display "Keyword ")
                  (display (substring (match:string m)
                                      (match:start m) 
                                      (match:end m)))
                  (display " is not implemented")
                  (newline)
                  (return #f)
                )
              )
            )
            (begin
              (display "ERROR")(newline)
              (return #f)
            )
          )
;          (display "rx-list")(display rx-list)(newline)
        )
        #t
      )
    )
  )
)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; public functions


(define (operation-e-lib-write)
  (case *SELECT_STATE*
    ( (cancel)
      (gui-entry-add-history "\n*Saving in e-lib format is cancelled*")
    )
    ( (string)
      (let*
        ( (filename (string-append *SELECT_ANSWER* ".mdf"))
          (fp (catch #t 
                     (lambda () (open-output-file filename))
                     (lambda args #f)))
        )
        (if fp
          (let* ( (point-map #f) )
            ; write the header
            (e-lib-write-header fp)
            ; write mesh points
            (set! point-map (e-lib-write-meshpoints fp))
            ; write elements
            (e-lib-write-elements point-map fp)
            ; close the file
            (close-output-port fp)
          )
          (gui-entry-add-history
            (string-append "\nERROR: File \"" filename 
                           "\" could not be opened for writing!")
          )
        )
      )
    )
  )
  ; clear keywords
  (selection-clear)
)



(define (operation-e-lib-read)
  (case *SELECT_STATE*
    ( (cancel)
      (gui-entry-add-history "\n*Opening cancelled*")
    )
    ( (string)
      (let* 
        ( (filename (string-append *SELECT_ANSWER* ".mdf"))
          (fp (catch #t 
                     (lambda () (open-input-file filename))
                     (lambda args #f)))
        )
        (if fp
          (let*
            ( (n 0) )
            ; reset objects
            (objects-clear-all)
            ; reset graphics
            (set! *ZOOM* 1.0)
            (set! *QUATERNION* '(0 0 0 1.0))
            (set! *XDIST* 0.0)
            (set! *YDIST* 0.0)
            
            ; read in the e-lib mesh file
            (if (not (e-lib-read-mesh fp))
              ; undo the operations
              (undo-all-operations)
            )
            (close-input-port fp)
            (set! n (+ (length *POINTS*)
                    (length *LINKS*)
                    (length *TRIANGS*)
                    (length *QUADS*)
                    (length *BOXES*)
                    (length *GROUPS*)))
            (if (= n 0)
              (set! n 1)
            )

            ; initialise screen selection
            (set! *GL_SELECT_BUFFER*
              (make-uniform-array 100 (* 5 n))
            )
            ; center of rotation is the center of the object
            (view-center-points *POINTS*)
            ; zoom to extension
            (view-zoom)
          )
          (gui-entry-add-history
            (string-append "\nERROR: File \"" filename 
                           "\" could not be opened for reading!")
          )
        )
      )
    )
  )
  ; clear keywords
  (selection-clear)
  ; update the screen
  (gui-update-gl-window #t)
)



