(define-module (sx-ext merge-nodes))
(use-modules  (sx selection-basic) (sx selection)
              (sx property) (sx basic) (sx undo)
              (sx gui-gl) (sx gui-entry) (sx gui-optmenu)
)


(export command-merge command-separate)


; add the commands to the sx system
(entry-add-command-list '("merge" "separate"))


(define (command-merge)
  ; setup selection mechanism
  (selection-setup selection-object
                   operation-merge
                   "Select points to merge:"
                   *OPTION_SELECT*
                   'select_obj
                   ; user must select points
                   '(no_empty type_point)
                   #f)
)


(define (command-separate)
  ; setup selection mechanism
  (selection-setup selection-object
                   operation-separate
                   "Select point:"
                   *OPTION_SELECT*
                   'select_obj
                   ; user must select point
                   '(no_empty no_multiple type_point)
                   #f)
)


(define (operation-merge)
  (case *SELECT_STATE*
    ( (objects)
      (if (>= (length *SELECT_LIST*) 2)
        (begin
          (set! *SELECT_STATE* 'init)
          (set! *SELECT_TEXT_PREFIX* "New position:")
          (set! *SELECT_INIT* (list 'no_empty))
          (set! *PICK_MODE* 'select_coord)
          (gui-entry-label-set-text *SELECT_TEXT_PREFIX*)
          (gui-entry-callback-replace selection-coord)
          (gui-option-menu-set *OPTION_POINT*)
        )
        (gui-entry-add-history " *More than two points are required*")
      )
    )
    ( (point)
      (let*
        ; let the user decide what he wants
        ( (prop-to-merge (object-property-collect *SELECT_LIST*))
          (new-prop '())
          (nn #f)
        )
        (if (not (null? prop-to-merge))
          (set! new-prop (gui-property-main-window prop-to-merge #t))
        )
        ; if the user accepted the edited properties
        (if new-prop
          (begin
            ; place the undo mark
            (undo-add-op-mark)
            ; create the new node
            (set! nn (point-add-basic
                       #f (list *SELECT_X* *SELECT_Y* *SELECT_Z*) #f #t))
            ; set the user properties
            (object-property-undo-update (list (list 'p nn)) new-prop #t #f)

            ; go through all elements connected to the points and
            ; replace the old node with the new one
            ; also set nodelem connectivity with new node
            (for-each 
              (lambda (x)
                ; all elements with this node
                (for-each
                  (lambda (y)
                    ; replace the node
                    (object-property-undo-set (car y) (cadr y) 'nodes
                      (map (lambda (z) (if (= z (cadr x)) nn z))
                        (object-property-get (car y) (cadr y) 'nodes)
                      )
                      #t #t
                    )
                  )
                  (object-property-get 'p (cadr x) 'parent)
                )
                ; add nodelem connectivity
                (object-property-undo-set 'p nn 'parent
                  (append (object-property-get 'p nn 'parent)
                          (object-property-get 'p (cadr x) 'parent) )
                  #t #t
                )
                ; no group related property should be in the point !
                ; delete the old node
                (object-property-undo-set 'p (cadr x) 'parent '() #t #t)
                (point-del-basic (cadr x) #t)
              )
              *SELECT_LIST*
            )
          )
        )
        ; just clear the selection
        (selection-clear)
        (gui-update-gl-window #t)
      )
    )
    ( (cancel)
      (selection-clear)
      (gui-update-gl-window #t)
    )
  )
)


(define (operation-separate)
  (case *SELECT_STATE*
    ( (objects)
      (if (= (length *SELECT_LIST*) 1)
        (let*
          ( (sel (car *SELECT_LIST*))
            (coord (object-property-get (car sel) (cadr sel) 'coords))
            (init-elems (object-property-get (car sel) (cadr sel) 'parent))
            (change-elems init-elems)
            (unmoveable-elems '())
          )
          ; check connected elements for child referencing
          ; go through all elements connected to the node
          (for-each
            (lambda (x)
              (let* 
                ( (child (object-property-get (car x) (cadr x) 'child)) )
                (if (not (null? child))
                  (begin
                    ; go through all elements in the children 
                    (for-each
                      (lambda (y)
                        (if (member y change-elems)
                          (begin
                            ; remove the child element from changing
                            ; and make it unmoveable
                            (set! change-elems (delete y change-elems))
                            (set! unmoveable-elems (cons y unmoveable-elems))
                          )
                        )
                      )
                      child
                    )
                    ; the parent itself is also unmoveable
                    (set! change-elems (delete x change-elems))
                    (set! unmoveable-elems (cons x unmoveable-elems))
                  )
                )
              )
            )
            init-elems
          )
          ; mark the beginning of the operation
          (undo-add-op-mark)
          ; go through all elements which need a new node
          (for-each
            (lambda (x)
              (let* 
                ( (cc '(0.0 0.0 0.0))
                  (nodes (object-property-get (car x) (cadr x) 'nodes))
                  (nn #f) 
                )
                ; calculate the centre of the element
                (for-each
                  (lambda (y)
                    (set! cc (map (lambda (xx yy) (+ xx yy))
                                  (object-property-get 'p y 'coords)
                                  cc)
                    )
                  )
                  nodes
                )
                ; move the new node towards the centre by 10 percent
                (set! cc (map-in-order (lambda (xx yy) 
                                (+ (* xx 0.8) (* 0.2 (/ yy (length nodes))))
                              )
                              coord cc))
                ; create the new point, preserve other properties as well
                (set! nn 
                      (point-add-basic #f cc 
                                       (object-property-get-all-user 'p 
                                                                     (cadr sel))
                                       #t)
                )
                ; set the nodelem connectivity of the point
                (object-property-undo-set 'p nn 'parent (list x) #t #t)
                ; change the node of the element
                (object-property-undo-set (car x) (cadr x) 'nodes
                  (map (lambda (z) (if (= z (cadr sel)) nn z))
                       (object-property-get (car x) (cadr x) 'nodes)
                  )
                  #t #t
                )
              )
            )
            change-elems
          )
          ; change the nodelem of the original node
          (object-property-undo-set (car sel) (cadr sel) 'parent 
                                    unmoveable-elems #t #t)

          ; clear selection and update screen
          (selection-clear)
          (gui-update-gl-window #t)
        )
        (gui-entry-add-history " *Only one point is required*")
      )
    )
    ( (cancel)
      (selection-clear)
      (gui-update-gl-window #t)
    )
  )
)








