(define-module (sx-ext surf-closed))
(use-modules  (sx selection-basic) (sx property) (sx undo) (sx basic)
              (sx gui-gl) (sx gui-entry) (sx gui-optmenu)
)

(export command-surf-closed)


; add the command to the sx system
(entry-add-command-list '("surf-closed"))

(define *OPTION_SURF_CLOSED* 
  (gui-option-menu-new
    '(("print" #t) ("single" #t) ("multiple" #t) ("all" #t))
  )
)


(define (command-surf-closed)
  (selection-setup selection-keyword
                   surf-closed-internal
                   "Mark [Print] sides with [Single/Multiple/All] adjacency <All> :"
                   *OPTION_SURF_CLOSED*
                   'view
                   #f
                   "Single Multiple All Print")
)




(define (surf-closed-internal)
  ; creates the intersection of two lists
  (define (set-intersection a b)
    (let* ( (is '()) )
      (for-each
        (lambda (x)
          (if (member x b)
            (set! is (cons x is))
          )
        )
        a
      )
      is
    )
  )

  (define (mark-boundary n1 n2 option)
    (case (length (set-intersection (object-property-get 'p n1 'parent)
                                    (object-property-get 'p n2 'parent)))
      ( (0) 
        (gui-entry-add-history (string-append "\nError on side"
                                              (number->string n1) " -- "
                                              (number->string n2) )
        )
        0
      )
      ( (1)
        (if (or (equal? option 'single)
                (equal? option 'all) )
          (begin
            (link-add-basic #f (list n1 n2) #f #t)
            1
          )
          (if (equal? option 'print)
            (begin
              (gui-entry-add-history (string-append "\n" 
                                                    (number->string n1) " -- "
                                                    (number->string n2)) )
              1
            )
            0
          )
        )
      )
      ( (2) 0 )
      ( else
        (if (or (equal? option 'multi)
                (equal? option 'all) )
          (begin
            (link-add-basic #f (list n1 n2) #f #t)
            1
          )
          (if (equal? option 'print)
            (begin
              (gui-entry-add-history (string-append "\n" 
                                                    (number->string n1) " -- "
                                                    (number->string n2)) )
              1
            )
            0
          )
        )
      )
    )
  )

  (let* ( (option 'all) (nlink 0) )
    (case *SELECT_STATE*
      ( (cancel) (set! option 'all))
      ( (keyword)
        (cond
          ( (string-ci=? *SELECT_ANSWER* "Single")
            (set! option 'single)
          )
          ( (string-ci=? *SELECT_ANSWER* "Multiple")
            (set! option 'multi)
          )
          ( (string-ci=? *SELECT_ANSWER* "All")
            (set! option 'all)
          )
          ( (string-ci=? *SELECT_ANSWER* "Print")
            (set! option 'print)
          )
        )
      )
    )
    ; place the undo mark
    (undo-add-op-mark)
    (for-each
      (lambda (triang)
        (let*
          ( (nodes (assoc-ref (cdr triang) 'nodes)) )
          (set! nlink (+ nlink 
                         (mark-boundary (car nodes)   (cadr nodes) option)))
          (set! nlink (+ nlink 
                         (mark-boundary (cadr nodes)  (caddr nodes) option)))
          (set! nlink (+ nlink 
                         (mark-boundary (caddr nodes) (car nodes) option)))
        )
      )
      *TRIANGS*
    )
    (for-each
      (lambda (quad)
        (let*
          ( (nodes (assoc-ref (cdr quad) 'nodes)) )
          (set! nlink (+ nlink 
                         (mark-boundary (car nodes)    (cadr nodes)   option)))
          (set! nlink (+ nlink 
                         (mark-boundary (cadr nodes)   (caddr nodes)  option)))
          (set! nlink (+ nlink 
                         (mark-boundary (caddr nodes)  (cadddr nodes) option)))
          (set! nlink (+ nlink 
                         (mark-boundary (cadddr nodes) (car nodes)    option)))
        )
      )
      *QUADS*
    )
    (gui-entry-add-history
            (string-append "\nNumber of sides: " (number->string nlink))
    )
  )

  (selection-clear)
  (gui-update-gl-window #t)
)


