(define-module (sx command))
(use-modules (sx selection-basic) (sx selection) (sx operation)
             (sx gui-optmenu) (sx view))
(export command-reset command-center command-style
        command-zoom command-light
        command-new command-open command-save
        command-quit command-point command-link 
        command-triang-point command-triang-side
        command-quad-point command-quad-side
        command-box 
        command-u command-redo command-undo
        command-getcoord command-getdist 
        command-select command-group command-ungroup
        command-delete command-move command-copy
        command-edit command-display
)

;;; -----------------------------------------------------------------
;;; High level functions to call by the user
;;; -----------------------------------------------------------------

(define (command-reset)
  (view-reset)
)


(define (command-center)
  (view-center)
)


(define (command-style)
  (view-style)
)

(define (command-zoom)
  (view-zoom)
)

(define (command-light)
  (view-light)
)


; -----------------------------------------------------------------


(define (command-new)
  (selection-setup selection-keyword
                   operation-new
                   "New model [Yes/No] <No>"
                   *OPTION_YES_NO* 'view #f "Yes No")
)

(define (command-open)
  (selection-setup selection-string
                   operation-open-native-format
                   "Enter filename (with path):"
                   *OPTION_EMPTY* 'view #f #f)
)

(define (command-save)
  (selection-setup selection-string
                   operation-save-native-format
                   "Enter filename (with path):"
                   *OPTION_EMPTY* 'view #f #f)
)

(define (command-quit)
  (operation-quit)
)

; -----------------------------------------------------------------

(define (command-point)
  ; setup selection mechanism
  (selection-setup selection-coord
                   operation-point-add
                   "Specify a point coordinate:" 
                   *OPTION_POINT*
                   'select_coord
                   #f #f)
)

(define (command-link)
  ; setup selection mechanism
  (selection-setup selection-object
                   operation-link-add
                   "Select link point(s):" 
                   *OPTION_SELECT*
                   'select_obj
                   ; user must select points
                   '(no_empty type_point no_multiple)
                   #f)
)

(define (command-triang-point)
  ; setup selection mechanism
  (selection-setup selection-object
                   operation-triang-point-add
                   "Select triangle point(s):" 
                   *OPTION_SELECT*
                   'select_obj
                   ; user must select points
                   '(no_empty type_point no_multiple)
                   #f)
)

(define (command-triang-side)
  ; setup selection mechanism
  (selection-setup selection-object
                   operation-triang-side-add
                   "Select triangle edge:" 
                   *OPTION_SELECT*
                   'select_obj
                   ; user must select link
                   '(no_empty type_link no_multiple)
                   #f)
)

(define (command-quad-point)
  ; setup selection mechanism
  (selection-setup selection-object
                   operation-quad-point-add
                   "Select quad point(s):" 
                   *OPTION_SELECT*
                   'select_obj
                   ; user must select points
                   '(no_empty type_point no_multiple)
                   #f)
)


(define (command-quad-side)
  ; setup selection mechanism
  (selection-setup selection-object
                   operation-quad-side-add
                   "Select quad edge:" 
                   *OPTION_SELECT*
                   'select_obj
                   ; user must select link
                   '(no_empty type_link no_multiple)
                   #f)
)

(define (command-box)
  ; setup selection mechanism
  (selection-setup selection-object
                   operation-box-add
                   "Select two box points:" 
                   *OPTION_SELECT*
                   'select_obj
                   ; user must select points
                   '(no_empty type_point no_multiple)
                   #f)
)


; -----------------------------------------------------------------

(define (command-u)
  (operation-u)
)

(define (command-redo)
  (operation-redo)
)

(define (command-undo)
  ; setup selection mechanism
  (selection-setup selection-integer
                   operation-undo
                   (string-append
                     "Number of operations to undo or "
                     "[Mark/undo All/undo Until mark] <1>" )
                   *OPTION_EMPTY*
                   'view
                   '(no_negative no_zero) 
                   "Until Mark All")
)

; -----------------------------------------------------------------
; query commands

(define (command-getcoord)
  ; setup selection mechanism
  (selection-setup selection-coord
                   operation-getcoord
                   "Select point:"
                   *OPTION_POINT*
                   'select_coord
                   #f #f)
)

(define (command-getdist)
  ; setup selection mechanism
  (selection-setup selection-object
                   operation-getdist
                   "Select two points:" 
                   *OPTION_SELECT*
                   'select_obj
                   ; user must select points
                   '(no_empty type_point)
                   #f)
)


; -----------------------------------------------------------------

(define (command-select)
  ; setup selection mechanism
  (selection-setup selection-object
                   operation-select 
                   "Select object(s):" 
                   *OPTION_SELECT*
                   'select_obj
                   '(type_any)
                   #f)
)


(define (command-group)
  ; setup selection mechanism
  (selection-setup selection-object
                   operation-group
                   "Select object(s):" 
                   *OPTION_SELECT*
                   'select_obj
                   ; user must select objects
                   '(no_empty type_any)
                   #f)
)


(define (command-ungroup)
  ; setup selection mechanism
  (selection-setup selection-object
                   operation-ungroup
                   "Select object(s):" 
                   *OPTION_SELECT*
                   'select_obj
                   '(type_group)
                   #f)
)


(define (command-delete)
  ; setup selection mechanism
  (selection-setup selection-object
                   operation-delete
                   "Select object(s):" 
                   *OPTION_SELECT*
                   'select_obj
                   '(type_any)
                   #f)

)


(define (command-move)
  ; setup selection mechanism
  (selection-setup selection-object
                   operation-move
                   "Select object(s):" 
                   *OPTION_SELECT*
                   'select_obj
                   '(type_any)
                   #f)
)


(define (command-copy)
  ; setup selection mechanism
  (selection-setup selection-object
                   operation-copy
                   "Select object(s):" 
                   *OPTION_SELECT*
                   'select_obj
                   '(type_any)
                   #f)
)



(define (command-edit)
  (selection-setup selection-object
                   operation-edit-property
                   "Select object(s):"
                   *OPTION_SELECT*
                   'select_obj
                   ; user must select point
                   '(type_any)
                   #f)
)

(define (command-display)
  (selection-setup selection-string
                   operation-display-property
                   "Name of property or ENTER to finish :"
                   *OPTION_EMPTY*
                   'view #f #f)
)

