(define-module (sx gui-main))
(use-modules (gtk gdk) (gtk gtk) (open-gl gtkgl)
             (sx gui-entry) (sx gui-gl) (sx gui-optmenu)
             (sx selection) (sx menu) (sx mouse))
(export gui-quit gui-main gui-create-main-window)



(define (gui-quit)
  (gtk-main-quit)
)

(define (gui-main)
  (gtk-main)
)

(define (gui-create-main-window)
  (define win-width	 #f)
  (define win-height	 #f)

  (let*
    ((window (gtk-window-new 'toplevel))
     (aspectframe (gtk-aspect-frame-new "3D view" 0.5 0.5 1.0 #f))
     (hbox     (gtk-hbox-new #f 0))
     (textbox  (gtk-hbox-new #f 0))
     (vbox     (gtk-vbox-new #f 0)) ; everything is packed into this
     (bottom_hbox  (gtk-hbox-new #f 0))
     (menu #f)
     (scrollbar #f)
    )
    ; this must be global
    (set! *OPENGL_WINDOW* (gtk-gl-area-new '(use-gl rgba doublebuffer) #f))
    (set! *ENTRY_LABEL*  (gtk-label-new *ENTRY_DEFAULT_TEXT*))
    (set! *ENTRY_TEXT* (gtk-entry-new))
    (set! *ENTRY_HISTORY* (gtk-text-new #f #f))
    (set! scrollbar (gtk-vscrollbar-new (gtk-text-vadj *ENTRY_HISTORY*)))
    (set! *OPTION_MENU* (gtk-frame-new "Option menu"))

    ; set my callback function
    (gui-entry-callback-replace gui-entry-default-callback)
    (gtk-signal-connect *ENTRY_TEXT* "activate" 
      (lambda ()
        (gui-entry-callback-execute)
      )
    )

    ; create optional menus
    (set! *OPTION_EMPTY* (gtk-vbox-new #f 0))
    (gtk-widget-show *OPTION_EMPTY*)
    (gui-create-option-menu)

    ; configure the main window properties 
    (gtk-window-set-title window " SX ")
    (gtk-container-border-width window 0)
    (gtk-signal-connect window "destroy" 
      (lambda ()
        (gtk-widget-destroy d)
        (gtk-main-quit)
      )
    )
    ; ignore window delete signals, exit only through quit command
    (gtk-signal-connect window "delete-event" (lambda (x) #t ))
    ; this is actually a gtk-signal-connect-after
    ; gui-entry-get-text is also modified for the space 
    (gtk-signal-connect window "key_press_event"
      (lambda (x) 
        (cond 
          ; ESC - cancel any operation
          ( (= (gdk-event-keyval x) 65307)
            (selection-cancel)
          )
          ; SPACE - execute the currently typed "command"
          ( (= (gdk-event-keyval x) 32)
            (gui-entry-callback-execute)
          )
        )
      )
      #f #t
    )

    ; pack things into the main window
    (gtk-container-add window vbox)
    ; menu creation
    (set! menu (load-menu "sx.mnu"))
    (if menu
      (gtk-box-pack-start vbox menu #f #f 3)
    )

    (gtk-container-add vbox hbox)
    (gtk-container-add hbox aspectframe)
    (gtk-container-add aspectframe *OPENGL_WINDOW*)

    ; option menu
    (gtk-box-pack-start hbox *OPTION_MENU* #f #f 3)
    (gui-option-menu-set *OPTION_EMPTY*)

    ; command area
    (gtk-widget-set-usize *ENTRY_HISTORY* 0 48)
    (gtk-box-pack-start textbox *ENTRY_HISTORY* #t #t 0)
    (gtk-box-pack-start textbox scrollbar #f #f 0)
    (gtk-box-pack-start vbox textbox #f #f 0)
    (gtk-container-border-width textbox 5)


    ; FIXME: not perfect, I need something as "clicked",
    ; the current setting leaves the mark around the text box
    (gtk-signal-connect *ENTRY_HISTORY* "focus_in_event"
      (lambda (x)
        ; keep focus at the entry
        (gtk-widget-grab-focus *ENTRY_TEXT*)
      )
    )

    (gtk-container-border-width bottom_hbox 5)
    (gtk-box-pack-start bottom_hbox *ENTRY_LABEL* #f #f 3)
    (gtk-box-pack-end bottom_hbox *ENTRY_TEXT* #t #t 3)

    (gtk-box-pack-end vbox bottom_hbox #f #f 0)

    ; configure the Open-GL area
    ;-----------------------------------------------------------------
    ; initial size
    (gtk-widget-set-usize *OPENGL_WINDOW* 500 500)
    ; select the events
    (gtk-widget-add-events *OPENGL_WINDOW* '(button-press-mask 
                                             button-motion-mask
                                             pointer-motion-mask
                                             button-release-mask))

    ; for expose
    (gtk-signal-connect *OPENGL_WINDOW* "expose_event"
      ; it only redraws the screen, no compilation is necessary
      (lambda (data) 
        (gui-update-gl-window #f)
      )
    )
    ; window resize change viewport of Open-GL area
    (gtk-signal-connect *OPENGL_WINDOW* "configure_event"
      (lambda (event)
        (gui-gl-viewport-set (gdk-event-configure-width  event)
                             (gdk-event-configure-height event))
      )
    )
    ; first time to show the Open-GL area
    (gtk-signal-connect *OPENGL_WINDOW* "realize"
      (lambda data (gui-init-gl-window))
    )
    ; a mouse button is pressed
    (gtk-signal-connect *OPENGL_WINDOW* "button_press_event"
      (lambda (event) 
        (GL-button-press-callback (gdk-event-x event) 
                                  (gdk-event-y event)
                                  (gdk-event-button event) )
      )
    )
    (gtk-signal-connect *OPENGL_WINDOW* "motion_notify_event"
      (lambda (event)
        (GL-button-motion-callback (gdk-event-x event) 
                                   (gdk-event-y event))
      )
    )
    (gtk-signal-connect *OPENGL_WINDOW* "button_release_event"
      (lambda (event) 
        (GL-button-release-callback (gdk-event-button event)
                                    (gdk-event-x event)
                                    (gdk-event-y event)
        )
      )
    )

    (gtk-widget-grab-focus *ENTRY_TEXT*)
    (gtk-widget-show-all window)
  )
)

