(define-module (sx gui-optmenu))
(use-modules (gtk gdk) (gtk gtk) (sx gui-entry))
(export *OPTION_MENU* *OPTION_EMPTY* 
        *OPTION_SELECT* *OPTION_POINT* *OPTION_YES_NO*
        gui-create-option-menu 
        gui-option-menu-new
        gui-option-menu-set gui-option-menu-get
)

(define *OPTION_MENU*    #f) ; the always visible part of option menu
(define *OPTION_EMPTY*   #f) ; an empty menu

(define *OPTION_SELECT*  #f) ; object selection optional menu
(define *OPTION_POINT*   #f) ; point specification optional menu
(define *OPTION_YES_NO*  #f) ; yes and no option

(define *OPTION_CURRENT* #f) ; the currently visible option menu


(define (gui-create-option-menu)
  ; selection optional menu
  ; these buttons activate the entry !!!
  (set! *OPTION_SELECT*
    (gui-option-menu-new 
      '(("last" #t) ("previous" #t) ("crossing" #t) ("all" #t)
        ("--" #t)
        ("single" #t) ("multiple" #t)
        ("--" #t) 
        ("only_point" #t)  ("only_link" #t) 
        ("only_triang" #t) ("only_quad" #t) ("only_box" #t) 
        ("only_group" #t)  ("any_object" #t)
        ("--" #t)
        ("view" #t) 
       )
    )
  )

  ; point optional menu
  ; these buttons only set the text in the entry
  (set! *OPTION_POINT* 
    (gui-option-menu-new 
      '((".xyz" #f) (".xy" #f) (".yz" #f) (".zx" #f) 
        (".x" #f) (".y" #f) (".z" #f)
        ("--" #t)
        ("view" #t)
       )
    )
  )

  (set! *OPTION_YES_NO*
    (gui-option-menu-new '(("yes" #t) ("no" #t)))
  )
)


;;; --------------------------------------------------------------
;;; GTK+ dependent part
;;; --------------------------------------------------------------

(define (gui-option-menu-new options)
  (let* ( (menu (gtk-vbox-new #f 0)) )
    (for-each
      (lambda (name)
        (if (string=? (car name) "--")
          (gtk-box-pack-start menu (gtk-label-new "---") #f #f 0)
          (let* ((button (gtk-button-new-with-label (car name))))
            (gtk-box-pack-start menu button #f #f 0)
            (gtk-signal-connect button "clicked" 
              (lambda ()
                ; keep focus at the entry
                (gtk-widget-grab-focus *ENTRY_TEXT*)
                
                (if (cadr name)
                  ; simply set the text ignoring current entry
                  ; and execute
                  (begin
                    (gui-entry-execute (car name))
;                    (gtk-entry-set-text *ENTRY_TEXT* (car name))
;                    (gtk-signal-emit *ENTRY_TEXT* "activate")
                  )
                  ; if it is not executed immediately
                  (gtk-entry-set-text *ENTRY_TEXT* 
                    (string-append (car name) ","))
                )
              )
            )
          )
        )
      )
      options
    )
    (gtk-widget-show-all menu)
    menu
  )
)


(define (gui-option-menu-set menu)
  ; change it only if it is a new
  (if (not (eq? menu *OPTION_CURRENT*))
    (begin
      ; remove the current 
      (if *OPTION_CURRENT*
        (gtk-container-remove *OPTION_MENU* *OPTION_CURRENT*)
      )
      ; add a new menu
      (gtk-container-add *OPTION_MENU* menu)
      (set! *OPTION_CURRENT* menu)
    )
  )
)

(define (gui-option-menu-get)
  *OPTION_CURRENT*
)



