(define-module (sx gui-property))
(use-modules (gtk gdk) (gtk gtk) (sx property))
(export gui-property-main-window)


;;; property editing
;;; --------------------------------------------------------------

(define (gui-property-main-window all-property no-multiple)
  (let*
    ( ; (all-property (object-property-collect elem-list))
      (window (gtk-window-new 'toplevel))
      (vbox   (gtk-vbox-new #f 0))
      (hbox   (gtk-hbox-new #f 0))

      (button (gtk-hbox-new #f 0))
      (ok     (gtk-button-new-with-label " Ok "))
      (cancel (gtk-button-new-with-label " Cancel "))

      (modify (gtk-vbox-new #f 0))
      (edit   (gtk-button-new-with-label " Edit "))
      (add    (gtk-button-new-with-label " Add "))
      (delete (gtk-button-new-with-label " Delete "))

      (table  (gtk-clist-new-with-titles '("Name" "Type" "Value")))

      (selected -1)

      (answer #f)
    )
    (gtk-signal-connect window "destroy" gtk-main-quit)
    (gtk-window-set-title window "Property editing")
    (gtk-window-set-position window 'mouse)


    ; set the size of the columns in the CList
    (gtk-clist-set-column-width table 0 100)
    (gtk-clist-set-column-width table 1 80)
    (gtk-clist-set-column-width table 2 200)

    ; write out the properties into the CList
    (for-each
      (lambda (y) (gtk-clist-append table y))
      (map
        (lambda (x)
          (list (car x)
                (if (> (length (cdr x)) 1)
                  "*multiple*"
                  (caadr x)
                )
                (if (or (> (length (cdr x)) 1)
                        (> (length (cdadr x)) 1) )
                  "*multiple*"
                  (cadadr x)
                )
          )
        )
        all-property
      )
    )

    ; store the selected row
    (gtk-signal-connect table "select_row"
      (lambda (row column event)
        (set! selected row)
      )
    )
    ; store that nothing is selected
    (gtk-signal-connect table "unselect_row"
      (lambda (row column event)
        (set! selected -1)
      )
    )

    ; editing a property
    (gtk-signal-connect edit "clicked"
      (lambda ()
        (let* 
          ( (lname '("")) 
            (name #f) (type #f) (value #f)
            (answer #f) 
          )
          ; if there is a selected row
          (if (> selected -1)
            (begin
              ; get the current values (name, type, value) from the CList
              (gtk-clist-get-text table selected 0 lname)
              (set! name (car lname))
              (gtk-clist-get-text table selected 1 lname)
              (set! type (car lname))
              (gtk-clist-get-text table selected 2 lname)
              (set! value (car lname))

              (set! answer 
                    (gui-property-edit-window #f name type value all-property))
              ; if it is a non-null list set the values in the CList
              (if (not (null? answer))
                (begin
                  (gtk-clist-set-text table selected 0 (car answer))
                  (gtk-clist-set-text table selected 1 (cadr answer))
                  (gtk-clist-set-text table selected 2 (caddr answer))
                )
              )
            )
          )
        )
      )
    )

    ; adding a new property
    (gtk-signal-connect add "clicked"
      (lambda ()
        ; create new property, no name or value is given
        (let* ((answer (gui-property-edit-window #t 
                                                 "*default name*" 
                                                 "word" 
                                                 "*value*" 
                                                 all-property)))
          ; if the user did set the value and the name
          (if (not (null? answer))
            ; if the property is already in the list, it cannot be added
            (if (assoc (car answer) all-property)
              (begin
                (gui-entry-add-history 
                  "\nERROR: The given property name alredy exists !")
                (gui-entry-add-history 
                  "\nERROR: The new property is not added !")
              )
              (begin
                ; append it to the CList
                (gtk-clist-append table answer)
                ; be sure that the all-property remembers it
                (set! all-property 
                      (acons (car answer) (list (cdr answer)) all-property))
              )
            )
          )
        )
      )
    )

    ; delete a property
    (gtk-signal-connect delete "clicked"
      (lambda ()
        (let* 
          ( (lname '(""))
            (name #f) 
          )
          (if (> selected -1)
            (begin
              ; get the current values
              (gtk-clist-get-text table selected 0 lname)
              (set! name (car lname))
              ; delete from the clist
              (gtk-clist-remove table selected)
              ; delete from exisiting all properties
              (set! all-property (assoc-remove! all-property name))
            )
          )
        )
      )
    )

    ; cancel property editing
    (gtk-signal-connect cancel "clicked"
      (lambda ()
        (gtk-widget-destroy window)
        (gtk-main-quit)
      )
    )

    ; assemble the new properties
    (gtk-signal-connect ok "clicked"
      (lambda ()
        (let* 
          ( (lname '(""))
            (name #f) (type #f) (value #f)
            (n (length all-property))
            (new-prop-list '())
          )
          (set! answer
            (call-with-current-continuation
              (lambda (return)
                ; collect all properties from the CList
                (do ((i 0 (1+ i))) ((= i n))
                  (gtk-clist-get-text table i 0 lname)
                  (set! name (car lname))
                  (gtk-clist-get-text table i 1 lname)
                  (set! type (car lname))
                  (gtk-clist-get-text table i 2 lname)
                  (set! value (car lname))
                  ; check whether we can have multiple definitions
                  (if (and no-multiple
                           (or (string=? value "*multiple*")
                               (string=? type  "*multiple*") ) )
                    ; if not error
                    (return #f)
                  )
                  ; add this property to the list
                  (set! new-prop-list
                        (cons (list (string->symbol name)
                                    (string->symbol type)
                                    (string->object-property 
                                      (string->symbol type) 
                                      value
                                    )
                              )
                              new-prop-list
                        )
                  )
                )
                new-prop-list
              )
            )
          )
          (if (not answer)
            (begin
              (gui-entry-add-history 
                "\nERROR: All *multiple* property must be resolved !")
            )
            (begin
              (gtk-widget-destroy window)
              (gtk-main-quit)
            )
          )
        )
      )
    )

    ; assemble the window
    (gtk-box-pack-start modify edit   #f #f 0)
    (gtk-container-border-width edit 5)
    (gtk-box-pack-start modify add    #f #f 0)
    (gtk-container-border-width add 5)
    (gtk-box-pack-start modify delete #f #f 0)
    (gtk-container-border-width delete 5)

    (gtk-container-add hbox  table)
    (gtk-container-border-width table 5)
    (gtk-box-pack-start hbox modify #f #f 0)

    (gtk-container-add vbox hbox)

    (gtk-box-pack-start button ok     #t #f 0)
    (gtk-box-pack-end   button cancel #t #f 0)
    (gtk-box-pack-start vbox button #f #f 0)
    (gtk-container-border-width button 5)

    (gtk-container-add window vbox)
    (gtk-container-border-width vbox 5)
    (gtk-widget-show-all window)

    ; grab the mouse for the window !!
    (gtk-grab-add window)
    ; start the event loop for this dialog
    (gtk-main)
    (gtk-grab-remove window)

    answer
  )
)

(define (gui-property-edit-window edit-name prop-name prop-type 
                                  prop-value all-property)
  (let*
    ( (value-choices (assoc prop-name all-property))
      (value-choices (if value-choices (cdr value-choices) '()))
      (window (gtk-window-new 'toplevel))
      (vbox   (gtk-vbox-new #f 0))
      (hbox   (gtk-hbox-new #f 0))

      (button (gtk-hbox-new #f 0))
      (ok     (gtk-button-new-with-label " Ok "))
      (cancel (gtk-button-new-with-label " Cancel "))

      (name   #f)
      (type   (gtk-combo-new))
      (value  (gtk-combo-new))

      (type-names '("word" "integer" "real" 
                    "word-list" "integer-list" "real-list"))

      (answer '())
    )
    (gtk-signal-connect window "destroy" gtk-main-quit)
    (gtk-window-set-title window "Property editing")
    (gtk-window-set-position window 'mouse)

    (if edit-name
      (begin
        (set! name (gtk-entry-new))
        (gtk-entry-set-text name prop-name)
        (gtk-entry-set-editable name edit-name)
      )
      (begin
        (set! name (gtk-label-new prop-name))
      )
    )

    ; types are not editable
    (gtk-entry-set-editable (gtk-combo-entry type) #f)
    ; set the type combo pop-up
    (if (> (length value-choices) 1)
      (set! type-names (cons "*multiple*" type-names))
    )
    (gtk-combo-set-popdown-strings type type-names)

    ; set the value combo pop-up
    (if (string=? prop-type "*multiple*")
      (begin
        ; when type is multiple, value is not editable
        (gtk-widget-set-sensitive value #f)
      )
      (let* ( (vv (assoc prop-type value-choices)) )
        (if vv
          (if (= (length (cdr vv)) 1)
            (gtk-combo-set-popdown-strings value (cdr vv))
            (gtk-combo-set-popdown-strings value (cons "*multiple*" (cdr vv)))
          )
          (gtk-combo-set-popdown-strings value '(""))
        )
      )
    )

    ; set the entry box to the current values 
    (gtk-entry-set-text (gtk-combo-entry type)  prop-type)
    (gtk-entry-set-text (gtk-combo-entry value) prop-value)

    ; when the type changes, change the available values
    (gtk-signal-connect (gtk-combo-entry type) "changed" 
      (lambda ()
        (let* ( (cur-type (gtk-entry-get-text (gtk-combo-entry type))) )
          (if (string=? cur-type "*multiple*")
            (begin
              (gtk-widget-set-sensitive value #f)
              (gtk-entry-set-text (gtk-combo-entry value) "*multiple*")
            )
            (let* ( (vv (assoc cur-type value-choices)) )
              ; enable editing the value entry
              (gtk-widget-set-sensitive value #t)
              (if vv
                (if (= (length (cdr vv)) 1)
                  (gtk-combo-set-popdown-strings value (cdr vv))
                  (gtk-combo-set-popdown-strings 
                    value 
                    (cons "*multiple*" (cdr vv))
                  )
                )
                (gtk-combo-set-popdown-strings value '(""))
              )
            )
          )
        )
      )
    )

    ; cancel editing the property
    (gtk-signal-connect cancel "clicked"
      (lambda ()
        (gtk-widget-destroy window)
        (gtk-main-quit)
      )
    )

    ; accept the edited property
    (gtk-signal-connect ok "clicked"
      (lambda ()
        (let* 
          ( (new-name  (if edit-name (gtk-entry-get-text name) prop-name))
            (new-type  (gtk-entry-get-text (gtk-combo-entry type)))
            (new-value (gtk-entry-get-text (gtk-combo-entry value)))
            ; try to convert it
            (new-prop  (string->object-property 
                         (string->symbol new-type) 
                         new-value)
            )
          )
          (if (and (or new-prop (string=? new-value "*multiple*"))
                   (> (string-length new-name) 0))
            (begin
              (set! answer (list new-name new-type new-value))
              (gtk-widget-destroy window)
              (gtk-main-quit)
            )
            (begin
              (if edit-name
                (begin
                  (gui-entry-add-history 
                    "\nERROR: Invalid property name or value !")
                )
                (begin
                  (gui-entry-add-history 
                    "\nERROR: Invalid property value !")
                )
              )
            )
          )
        )
      )
    )

    ; assemble the window
    (gtk-container-add hbox name)
    (gtk-container-add hbox type)
    (gtk-container-border-width type 5)
    (gtk-container-add hbox value)
    (gtk-container-border-width type 5)
    (gtk-container-add vbox hbox)

    (gtk-box-pack-start button ok     #t #f 0)
    (gtk-box-pack-end   button cancel #t #f 0)
    (gtk-box-pack-start vbox button #f #f 0)
    (gtk-container-border-width button 5)

    (gtk-container-add window vbox)
    (gtk-container-border-width vbox 5)
    (gtk-widget-show-all window)
    
    (gtk-grab-add window)
    (gtk-main)
    (gtk-grab-remove window)

    answer
  )
)


