(define-module (sx gui-selection))
(use-modules (gtk gdk) (gtk gtk) 
             (sx property) (sx basic) (sx selection-basic) (sx view)
             (sx gui-gl))
(export gui-selection-single-object gui-selection-single-coord)

; selects a single point and returns the coordinate of the point or #f
(define (gui-selection-single-coord sel)
  (if sel
    (let* 
      ( (plist '())
        (point '())
      )
      ; collect only points from the selection
      (for-each 
        (lambda (x)
          (if (eq? (car x) 'p)
            (set! plist (cons x plist))
          )
        )
        sel
      )
      ; do a single selection on it
      (if (not (null? plist))
        (set! point (gui-selection-single-object plist))
      )
      ; return the coordinates or false
      (if (not (null? point))
        (object-property-get 'p (cadar point) 'coords)
        #f
      )
    )
    #f
  )
)




; this function allows the selection of only one object from the list
; it returns an empty list in the case "cancel" or a list with the
; selected element
; The function uses the GUI to display the options
(define (gui-selection-single-object olist)
  (let*
    ( (object #f) )
    (cond
      ( (> (length olist) 1) ; several objects are selected
        ; push the current selection on the selection stack
        (selection-push-internal)
        ; switch to viewing mode without an exit
        (show-setup)
        (gui-update-gl-window #t)
        ; pop up window for selection
        (set! object (gui-selection-single-object-dialog olist))
        ; restore the actual selection
        (selection-pop-internal)
        ; restore the screen
        (gui-update-gl-window #t)
      )
      ( (= (length olist) 1) ; only one object is selected
        (set! object (car olist)) )
    )
    ; validate the object as well
    (if (and object (object-valid object))
      (list object)
      '()
    )
  )
)


;;; ---------------------------------------------------------------------------
;;; GTK+ dependent part
;;; ---------------------------------------------------------------------------

(define (gui-selection-single-object-dialog plist)
  (let*
    ((pos 0)
     (d (gtk-window-new 'dialog))
     (v (gtk-vbox-new #f 0))
     (h (gtk-hbox-new #t 10))
     (l1 (gtk-label-new "You have selected multiple objects!"))
     (l2 (gtk-label-new "Which ONE do you want?"))
     (l (gtk-label-new (object-property-print (list-ref plist pos))))
     (s (gtk-hseparator-new))
     (a (gtk-button-new-with-label " Select "))
     (p (gtk-button-new-with-label " << "))
     (n (gtk-button-new-with-label " >> "))
     (c (gtk-button-new-with-label " Cancel "))
     (show (gtk-button-new-with-label " Show "))
     (answer #f)
    )

    (gtk-window-set-title d "Object selection")
    (gtk-window-set-position d 'mouse)

    (gtk-container-border-width h 5)
    (gtk-widget-set-usize l1 100 20)
    (gtk-widget-set-usize l2 100 20)
    (gtk-widget-set-usize l 100 30)
    (gtk-container-add d v)
    (gtk-box-pack-start v l1 #f #f 0)
    (gtk-box-pack-start v l2 #f #f 0)
    (gtk-box-pack-start v l #f #f 0)
    (gtk-box-pack-start v s #f #f 0)
    (gtk-box-pack-start v h #f #f 0)
    (gtk-box-pack-start h a #f #t 0)
    (gtk-box-pack-start h p #f #t 0)
    (gtk-box-pack-start h n #f #t 0)
    (gtk-box-pack-start h c #f #t 0)
    (gtk-box-pack-start h show #f #t 0)

    ; window is killed
    (gtk-signal-connect d "delete_event" 
      (lambda (event)
        (gtk-widget-destroy d)
        (gtk-main-quit)
        #t
      )
    )

    ; when the object is selected
    (gtk-signal-connect a "clicked" 
      (lambda () 
        (set! answer (list-ref plist pos))
        (gtk-widget-destroy d)
        (gtk-main-quit)
      )
    )
    ; previous object
    (gtk-signal-connect p "clicked" 
      (lambda ()
        (if (= pos 0)
          (set! pos (1- (length plist)))
          (set! pos (1- pos))
        )
        (gtk-label-set-text l (object-property-print (list-ref plist pos)))
      )
    )
    ; next object
    (gtk-signal-connect n "clicked" 
      (lambda ()
        (if (= pos (1- (length plist)))
          (set! pos 0)
          (set! pos (1+ pos))
        )
        (gtk-label-set-text l (object-property-print (list-ref plist pos)))
      )
    )
    ; cancel the selection
    (gtk-signal-connect c "clicked" 
      (lambda ()
        (gtk-widget-destroy d)
        (gtk-main-quit)
      )
    )
    ; show the element
    (gtk-signal-connect show "clicked" 
      (lambda ()
        (set! *SELECT_LIST* (list (list-ref plist pos)))
        (gui-update-gl-window #t)
      )
    )

    (gtk-widget-show-all d)
    ; register the dialog window
    (set! *DIALOG_WINDOWS* (cons d *DIALOG_WINDOWS*))
    (gtk-main)
    (set! *DIALOG_WINDOWS* (cdr *DIALOG_WINDOWS*))

    answer
  )
)
