(define-module (sx menu))
(use-modules (ice-9 rdelim) 
             (sx gui-menu)
             (sx gui-entry)
)
(export load-menu)


; menu function
; --------------------------------------------------------------------

;(define-module (sx menu))
;(export load-menu)

(define *LINE_NUMBER* 0)

(define (print-line-error text)
  (gui-entry-add-history (string-append "\nERROR: "
                                        text
                                        " at line "
                                        (number->string *LINE_NUMBER*)
                                        "\n"))
)


; it reads a valid line from the file, it also does basic validation
; and converts the line to a list
; it can return:
; eof - at the end of file
; error - problem with line
; list - the list of read expression
(define (read-menu-line fp)
  (call-with-current-continuation
    (lambda (return)
      (let*
        ((line #f) (comma 0) (ll 0))
        (do () (#f)
          (set! line (read-line fp))
          (set! *LINE_NUMBER* (1+ *LINE_NUMBER*))
          (if (eof-object? line) (return 'eof))

          (set! ll (string-length line))
          ; if there is a string, check the number of double quotes
          (if (> ll 0)
            (begin
              (set! comma 0)
              (do ((i 0 (1+ i))) ((= i ll))
                (if (eq? (string-ref line i) #\")
                  (set! comma (1+ comma))
                )
              )
            )
          )
          (if (= (modulo comma 2) 1)
            (begin
              (print-line-error "Double quotes are not balanced")
              (return 'error)
            )
          )
          (set! line (call-with-input-string (string-append "(" line ")") read))
          (if (not (null? line))
            (return line)
          )
        )
        'error
      )
    )
  )
)


; returns #t or #f
(define (menu-parse label fp)
  (call-with-current-continuation
    (lambda (return)
      (let*
        ((expr #f) (item #f)
         (menu (gui-menu-new)) ; create a menu
         (menu_head (gui-menu-item-new label)) ; create the head of the menu
        )
        (gui-submenu-set menu menu_head) ; add the menu to the head
        (do () (#f)
          (set! expr (read-menu-line fp))
          (case expr
            ((error eof) (return #f))
            (else
              (if (eq? (car expr) '*ENDMENU)
                (return menu_head)
              )
              (if (not (string? (car expr)))
                (begin
                  (print-line-error "Menu item name must be a string")
                  (return #f)
                )
              )
              (if (string=? (car expr) "--")
                (gui-menu-append menu (gui-menu-separator-new))
                (begin
                  (set! item (gui-menu-item-new (car expr)))
                  (gui-menu-append menu item)
                )
              )
              (if (= (length expr) 2)
                (gui-menu-item-connect item (text-split-with-empty (cadr expr) 
                                                                   #\space 
                                                                   #t))
              )
            )
          )
        )
        (print-line-error "internal error")
        #f
      )
    )
  )
)


(define (load-menu file)
  (let*
    ((fp (catch #t (lambda () (open-input-file file))
                   (lambda args #f)))
     (menubar (gui-menu-bar-new))
     (menu #f)
     (exitval #f)
    )
    (set! *LINE_NUMBER* 0)

    (if fp
      (begin
        (set! exitval
          (call-with-current-continuation
           (lambda (return)
              (let* 
                ((line #f))
                (do () (#f)
                  (set! line (read-menu-line fp))
                  (case line
                    ((eof) (return menubar))
                    ((error) (return #f))
                    (else
                      (if (and (= (length line) 2)
                               (equal? (car line) '*MENU))
                        (begin
                          (if (not (string? (cadr line)))
                            (begin
                              (print-line-error "Menu name must be a string")
                              (return #f)
                            )
                          )
                          (set! menu (menu-parse (cadr line) fp))
                          (if menu
                            (gui-menu-bar-append menubar menu)
                            (return #f)
                          )
                        )
                        (begin
                          (print-line-error "Cannot interpret expression")
                          (return #f)
                        )
                      )
                    )
                  )
                )
                #t
              )
            )
          )
        )
        ; close the file
        (close-input-port fp)
      )
    )
    ; return the indicator of success or failure
    exitval
  )
)



