(define-module (sx operation))
(use-modules (ice-9 rdelim)
             (sx selection-basic) (sx selection) (sx view)
             (sx undo) (sx property) (sx basic) (sx render)
             (sx gui-main) (sx gui-optmenu) (sx gui-entry) (sx gui-gl)
             (sx gui-property)
)
(export operation-new operation-open-native-format operation-save-native-format
        operation-quit
        operation-u operation-redo operation-undo
        operation-point-add operation-link-add
        operation-triang-point-add operation-triang-side-add
        operation-quad-point-add operation-quad-side-add
        operation-box-add operation-group operation-ungroup
        operation-select operation-delete operation-move operation-copy
        operation-edit-property operation-display-property
        operation-getcoord operation-getdist
)


;---------------------------------------------------------
; Functions operating on the model in one way or another
;
; If an operation is undoable it MUST put down an
; OPERATION-MARK at the beginning of the operation, so
; the undo system will be able to undo it !!!
;---------------------------------------------------------



(define (operation-new)
  (case *SELECT_STATE*
    ( (cancel)
      (gui-entry-add-history "\n*New model creation cancelled*")
    )
    ( (keyword)
      (cond
        ( (string-ci=? *SELECT_ANSWER* "Yes")
          ; remove all objects
          (objects-clear-all)

          ; reset graphics
          (set! *ZOOM* 1.0)
          (set! *QUATERNION* '(0 0 0 1.0))
          (set! *XDIST* 0.0)
          (set! *YDIST* 0.0)
          ; reset the rotation centre to origo
          (set! *GL_CENTER* '(0.0 0.0 0.0))

          ; it looks as it has been saved :-)
          (set! *CHANGED* #f)
          ; update the screen
          (gui-update-gl-window #t)
        )
        ( (string-ci=? *SELECT_ANSWER* "No")
          (gui-entry-add-history "\n*New model creation cancelled*")
        )
      )
    )
  )
  ; clear keywords
  (selection-clear)
)


(define (operation-open-native-format)
  (case *SELECT_STATE*
    ( (cancel)
      ; default option
      (gui-entry-add-history "\n*Opening cancelled*")
    )
    ( (string)
      (let* ((fp (catch #t (lambda () (open-input-file *SELECT_ANSWER*))
                           (lambda args #f))) )
        (if fp
          (begin
            (close-input-port fp)
            ; reset objects
            (objects-clear-all)
            ; reset graphics
            (set! *ZOOM* 1.0)
            (set! *QUATERNION* '(0 0 0 1.0))
            (set! *XDIST* 0.0)
            (set! *YDIST* 0.0)
            ; load the file
            ; FIXME: catch errors !!!
            (load *SELECT_ANSWER*)
            ; initialise screen selection
            (set! *GL_SELECT_BUFFER*
              (make-uniform-array 100 (* 5 (+ (length *POINTS*)
                                              (length *LINKS*)
                                              (length *TRIANGS*)
                                              (length *QUADS*)
                                              (length *BOXES*)
                                              (length *GROUPS*)))
              )
            )
            ; it looks as it has been saved :-)
            (set! *CHANGED* #f)
          )
          (gui-entry-add-history
            (string-append "\nERROR: File \"" 
                           *SELECT_ANSWER* 
                           "\" cannot be opened for reading!")
          )
        )
      )
    )
  )
  ; clear keywords
  (selection-clear)
  ; update the screen
  (gui-update-gl-window #t)
)


(define (operation-save-native-format)
  (case *SELECT_STATE*
    ( (cancel)
      ; default option
      (gui-entry-add-history "\n*Saving cancelled*")
    )
    ( (string)
      (let*
        ( (fp (catch #t (lambda () (open-output-file *SELECT_ANSWER*))
                                   (lambda args #f))) )
        (if fp
          (begin
            ; object database
            (write-line (list 'set! '*POINTS* 
                              (list 'quote *POINTS*)) fp)
            (write-line (list 'set! '*LINKS* 
                              (list 'quote *LINKS*)) fp)
            (write-line (list 'set! '*TRIANGS* 
                              (list 'quote *TRIANGS*)) fp)
            (write-line (list 'set! '*QUADS* 
                              (list 'quote *QUADS*)) fp)
            (write-line (list 'set! '*BOXES* 
                              (list 'quote *BOXES*)) fp)
            (write-line (list 'set! '*GROUPS* 
                              (list 'quote *GROUPS*)) fp)

            (write-line (list 'set! '*POINT_UNIQUE*  *POINT_UNIQUE*)  fp)
            (write-line (list 'set! '*LINK_UNIQUE*   *LINK_UNIQUE*)   fp)
            (write-line (list 'set! '*TRIANG_UNIQUE* *TRIANG_UNIQUE*) fp)
            (write-line (list 'set! '*QUAD_UNIQUE*   *QUAD_UNIQUE*)   fp)
            (write-line (list 'set! '*BOX_UNIQUE*    *BOX_UNIQUE*)   fp)
            (write-line (list 'set! '*GROUP_UNIQUE*  *GROUP_UNIQUE*)  fp)

            (write-line (list 'set! '*POINT_MAX* *POINT_MAX*) fp)

            ; viewing variables
            (write-line (list 'set! '*ZOOM* *ZOOM*) fp)
            (write-line (list 'set! '*QUATERNION* 
                              (list 'quote *QUATERNION*)) fp)
            (write-line (list 'set! '*XDIST* *XDIST*) fp)
            (write-line (list 'set! '*YDIST* *YDIST*) fp)

            (write-line (list 'set! '*GL_CENTER* 
                              (list 'quote *GL_CENTER*)) fp)

            (close-output-port fp)
            ; mark, that everything has been recorded to a file
            ; it is safe to quit
            (set! *CHANGED* #f)
          )
          (gui-entry-add-history "\nERROR: Could not open file for writing !")
        )
      )
    )
  )
  ; clear keywords
  (selection-clear)
)

(define (operation-quit)
  (if *CHANGED*
    (selection-setup selection-keyword
                     operation-quit-internal
                     "Model has changed, quit anyway [Yes/no] <No>"
                     *OPTION_YES_NO* 'view #f "Yes No")
    (gui-quit)
  )
)


(define (operation-quit-internal)
  (case *SELECT_STATE*
    ( (cancel)
      (gui-entry-add-history "\n*Quitting cancelled*")
    )
    ( (keyword)
      (cond
        ( (string-ci=? *SELECT_ANSWER* "Yes")
          (gui-quit)
        )
        ( (string-ci=? *SELECT_ANSWER* "No")
          (gui-entry-add-history "\n*Quitting cancelled*")
        )
      )
    )
  )
  ; clear keywords
  (selection-clear)
)

;---------------------------------------------------------

(define (operation-u)
;  (display (undo-get-list))(newline)

  (if (= (undo-get-size) 0)
    (gui-entry-add-history "\n*No more undo*")
    (begin
      (undo-one-operation)
      (gui-update-gl-window #t)
    )
  )
;  (display (undo-get-list))(newline)
)


(define (operation-redo)
  (redo-operation)
  (gui-update-gl-window #t)
)

(define (operation-undo)
  (case *SELECT_STATE*
    ( (cancel)  (operation-u) )
    ( (integer) 
      (do ((i *SELECT_ANSWER* (1- i))) ((or (= i 0) (= (undo-get-size) 0)))
        (operation-u)
      )
    )
    ( (keyword)
      (cond
        ( (string-ci=? *SELECT_ANSWER* "Mark")
          (begin
            (undo-add-user-mark)
            (gui-entry-add-history "\n*Mark added to undo list*")
          )
        )
        ( (string-ci=? *SELECT_ANSWER* "Until")
          (if (= (undo-get-size) 0)
            (gui-entry-add-history "\n*No more undo*")
            (begin
              (undo-until-user-mark)
            )
          )
        )
        ( (string-ci=? *SELECT_ANSWER* "All")
          (if (= (undo-get-size) 0)
            (gui-entry-add-history "\n*No more undo*")
            (begin
              (undo-all-operations)
            )
          )
        )
      )
    )
  )

  ; clear keywords, integer selection
  (selection-clear)
  ; update the screen, 
  ; something has changed in the object database !
  (gui-update-gl-window #t)
)


;---------------------------------------------------------


; input is a series of link elements: '((l 71) (l 13) (l 45))
; returns #f in the case of an error or the list of nodes which
; defines the closed polygon (3 4 5 6)
(define (closed-polygon? selection)
  (call-with-current-continuation
    (lambda (return)
      (let* 
        ( (links (map (lambda (x) (cadr x)) selection))
          (node-list (map (lambda (x) 
                            (object-property-get (car x) (cadr x) 'nodes))
                          selection))
          (nodes (car node-list))
          (node-list (cdr node-list))
          (elem #f)
        )
        ; validate that the links make up a closed polygon
        (while (not (null? node-list))
          (let loop ( (x node-list) )
            (if (null? x)
              (begin
                (gui-entry-add-history " *Not a closed polygon definition*")
                (return #f)
              )
            )
            (set! elem (car x))
            (if (equal? (car elem) (car nodes))
              (begin
                (set! nodes (append (cdr elem) nodes))
                (set! node-list (delete elem node-list))
              )
              (if (equal? (cadr elem) (car nodes))
                (begin
                  (set! nodes (cons (car elem) nodes))
                  (set! node-list (delete elem node-list))
                )
                (loop (cdr x))
              )
            )
          )
        )
        (let* 
          ( (l-nodes (reverse nodes))
            (first   (car l-nodes))
            (l-nodes (reverse (cdr l-nodes)))
          )
          (if (not (equal? first (car l-nodes)))
            (begin
              (gui-entry-add-history " *Not a closed polygon definition*")
              (return #f)
            )
          )
          l-nodes
        )
      )
    )
  )
)


(define (operation-point-add)
  (case *SELECT_STATE*
    ((point)
      (undo-add-op-mark)
      ; add the point with the coordinates
      (point-add-basic #f (list *SELECT_X* *SELECT_Y* *SELECT_Z*) #f #t)
      ; clear it for new points
      (set! *SELECT_X* #f)
      (set! *SELECT_Y* #f)
      (set! *SELECT_Z* #f)
      ; reset the text (for example .xy was used)
      (gui-entry-label-set-text *SELECT_TEXT_PREFIX*)
      ; update the screen, because object database has changed
      (gui-update-gl-window #t)
    )
    ((cancel) ; stop adding points
      ; clear all selection variables
      (selection-clear)
    )
  )
)


(define (operation-link-add)
  (if (= (length *SELECT_LIST*) 2)
    (begin
      (undo-add-op-mark)
      ; add the link
      (link-add-basic #f (list (cadar *SELECT_LIST*) 
                               (cadadr *SELECT_LIST*)) #f #t)

      ; clear all selection
      (selection-clear)
      ; update the screen
      (gui-update-gl-window #t)
    )
    (gui-entry-add-history " *Two points are required*")
  )
)


(define (operation-triang-point-add)
  (if (= (length *SELECT_LIST*) 3)
    (let* 
      ( (nodes (reverse (map (lambda (x) (cadr x)) *SELECT_LIST*))) )
      (undo-add-op-mark)
      ; add the triangle
      (triang-add-basic #f nodes #f #t)

      ; clear all selection
      (selection-clear)
      ; update the screen
      (gui-update-gl-window #t)
    )
    (gui-entry-add-history " *Three points are required*")
  )
)


(define (operation-triang-side-add)
  (if (= (length *SELECT_LIST*) 3)
    (let* 
      ( (nodes (closed-polygon? *SELECT_LIST*)) )
      (if nodes
        (let*
          ( (prop (list (cons 'child *SELECT_LIST*))) )
          (undo-add-op-mark)
          (triang-add-basic #f nodes prop #t)
        )
      )
      ; clear all selection
      (selection-clear)
      ; update the screen
      (gui-update-gl-window #t)
    )
    (gui-entry-add-history " *Three link elements are required*")
  )
)


(define (operation-quad-point-add)
  (if (= (length *SELECT_LIST*) 4)
    (let* 
      ( (nodes (reverse (map (lambda (x) (cadr x)) *SELECT_LIST*))) )
      (undo-add-op-mark)
      ; add the quad
      (quad-add-basic #f nodes #f #t)

      ; clear all selection
      (selection-clear)
      ; update the screen
      (gui-update-gl-window #t)
    )
    (gui-entry-add-history " Four points are required*")
  )
)


(define (operation-quad-side-add)
  (if (= (length *SELECT_LIST*) 4)
    (let* 
      ( (nodes (closed-polygon? *SELECT_LIST*)) )
      (if nodes
        (let*
          ( (prop (list (cons 'child *SELECT_LIST*))) )
          (undo-add-op-mark)
          (quad-add-basic #f nodes prop #t)
        )
      )
      ; clear all selection
      (selection-clear)
      ; update the screen
      (gui-update-gl-window #t)
    )
    (gui-entry-add-history " *Four link elements are required*")
  )
)


(define (operation-box-add)
  (if (= (length *SELECT_LIST*) 2)
    (begin
      (undo-add-op-mark)
      ; add the box
      (box-add-basic #f (list (cadar *SELECT_LIST*) 
                              (cadadr *SELECT_LIST*)) #f #t)

      ; clear all selection
      (selection-clear)
      ; update the screen
      (gui-update-gl-window #t)
    )
    (gui-entry-add-history " *Two points are required*")
  )
)



(define (operation-select)
  ; clear all selection
  (selection-clear)
  ; only update the screen 
  (gui-update-gl-window #t)
)



(define (operation-group)
  (if (not (null? *SELECT_LIST*))
    (begin
      (undo-add-op-mark)
      ; ensure that all nodes that are referenced by elements are
      ; added to the group, but do not expand groups itself
      (selection-add-ref-nodes #f)
      ; group them
      (group-create-basic #f *SELECT_LIST* #f #t)
    )
  )
  ; clear all selection
  (selection-clear)
  ; update the screen
  (gui-update-gl-window #t)
)


(define (operation-ungroup)
  ; mark the operation
  (undo-add-op-mark)
  (if (not (null? *SELECT_LIST*))
    (for-each 
      (lambda (x) (if (eq? (car x) 'g) (group-explode-basic (cadr x) #t)))
      *SELECT_LIST*
    )
  )
  ; clear all selection
  (selection-clear)
  ; update the screen
  (gui-update-gl-window #t)
)


(define (operation-delete)
  (if (not (null? *SELECT_LIST*))
    (begin
      (undo-add-op-mark)
      ; first delete the elements
      (for-each (lambda (x) 
          (case (car x)
            ((l) (link-del-basic   (cadr x) #t))
            ((t) (triang-del-basic (cadr x) #t))
            ((q) (quad-del-basic   (cadr x) #t))
            ((b) (box-del-basic    (cadr x) #t))
            ((g) (gui-entry-add-history 
                   "\nWARNING: Groups must be exploded before deleting !")
            )
          )
        )
        *SELECT_LIST*
      )
      ; the delete the points 
      (for-each (lambda (x) 
        (if (eq? (car x) 'p) (point-del-basic (cadr x) #t)))
        *SELECT_LIST*
      )
    )
  )

  ; clear all selection
  (selection-clear)
  ; update the screen
  (gui-update-gl-window #t)
)


(define (object-move-internal)
  ; add referenced points
  ; expand to nodes enclosed by groups
  (selection-add-ref-nodes #t)

  ; move the points
  (for-each 
    (lambda (x)
      (if (eq? (car x) 'p)
        (let* 
          ((n (cadr x))
           (co (object-property-get 'p n 'coords))
           (xcoord (+ (car co) *SELECT_X*))
           (ycoord (+ (cadr co) *SELECT_Y*))
           (zcoord (+ (caddr co) *SELECT_Z*))
          )
          ; change the position of the point
          (object-property-undo-set 
            'p n 'coords (list xcoord ycoord zcoord) #t #t)
        )
      )
    )
    *SELECT_LIST*
  )
)


; objects are selected, now we need a displacement or two points
(define (operation-move)
  (if (null? *SELECT_LIST*)
    (set! *SELECT_STATE* 'quit)
  )
  (case *SELECT_STATE*
    ( (objects)
      ; change the text
      (set! *SELECT_TEXT_PREFIX* 
            "Select base point or displacement vector:")
      ; user must select distance or two points
      (set! *SELECT_INIT* (cons 'no_empty *SELECT_INIT*))
      (set! *PICK_MODE* 'select_coord)
      (gui-entry-label-set-text *SELECT_TEXT_PREFIX*)
      ; stop the signal as well
      (gui-entry-callback-replace selection-dist-point)
      (gui-option-menu-set *OPTION_POINT*)
    )
    ( (dist)
      ; write back the displacement vector
      (gui-entry-add-history 
        (string-append "\nDistance vector: "
                       (number->string *SELECT_X*) " "
                       (number->string *SELECT_Y*) " "
                       (number->string *SELECT_Z*)))

      ; mark the operation 
      (undo-add-op-mark)
      ; move the objects
      (object-move-internal)

      ; clear all selection variables
      (selection-clear)
      ; update the screen
      (gui-update-gl-window #t)
    )
    ( (cancel quit)
      ; clear all selection variables
      (selection-clear)
      ; update the screen
      ; removes selection coloring
      (gui-update-gl-window #t)
    )
  )
)


; returns with the created elements type and id in a list
; it can be called recursively
(define (object-duplicate-internal type id node_assoc)
  (let* 
    ( (nodes (object-property-get type id 'nodes))
      (elems '())
    )
    (case type
      ( (l)
        ; map the old nodes to the new ones
        (set! nodes (map (lambda (x) (car (assoc-ref node_assoc x))) nodes))
        (list 'l (link-add-basic #f nodes (object-property-get-all-user 'l id) #t))
      )
      ( (t)
        ; map the old nodes to the new ones
        (set! nodes (map (lambda (x) (car (assoc-ref node_assoc x))) nodes))
        (list 't (triang-add-basic #f nodes (object-property-get-all-user 't id) #t))
      )
      ( (q)
        ; map the old nodes to the new ones
        (set! nodes (map (lambda (x) (car (assoc-ref node_assoc x))) nodes))
        (list 'q (quad-add-basic #f nodes (object-property-get-all-user 'q id) #t))
      )
      ( (b)
        ; map the old nodes to the new ones
        (set! nodes (map (lambda (x) (car (assoc-ref node_assoc x))) nodes))
        (list 'b (box-add-basic #f nodes (object-property-get-all-user 'b id) #t))
      )
      ( (g)
        (for-each
          (lambda (x)
            (case (car x)
              ( (p)
                (set! elems 
                      (cons (list 'p (car (assoc-ref node_assoc (cadr x)))) elems) )
              )
              ( else
                (set! elems 
                      (cons (object-duplicate-internal (car x)
                                                       (cadr x)
                                                       node_assoc) elems)
                )
              )
            )
          )
          (object-property-get 'g id 'parts)
        )
        (list 'g (group-create-basic #f elems (object-property-get-all-user 'g id) #t))
      )
    )
  )
)


(define (object-copy-internal)
  (let*
    ( (nodes #f) (nplist '()) )
    ; add referenced points
    (selection-add-ref-nodes #t)

    ; create and add points
    (for-each
      (lambda (x)
        (if (eq? (car x) 'p)
          (let* 
            ( (n (cadr x))
              (co (object-property-get 'p n 'coords))
              (xcoord (+ (car co) *SELECT_X*))
              (ycoord (+ (cadr co) *SELECT_Y*))
              (zcoord (+ (caddr co) *SELECT_Z*))
            )
            ; add the new point
            (set! nplist (cons (list n (point-add-basic ;create new point
                                        #f 
                                        (list xcoord ycoord zcoord)
                                        (object-property-get-all-user 'p n)
                                        #t)
                               )
                               nplist) )
          )
        )
      )
      *SELECT_LIST*
    )
    ; add all new element
    (for-each
      (lambda (x) (object-duplicate-internal (car x) (cadr x) nplist) )
      *SELECT_LIST*
    )

  )
)


; objects are selected, now we need a displacement or two points
(define (operation-copy)
  (if (null? *SELECT_LIST*)
    (set! *SELECT_STATE* 'quit)
  )
  (case *SELECT_STATE*
    ((objects)
         ; change the text
         (set! *SELECT_TEXT_PREFIX* "Select base point or displacement vector:")
         ; user must select distance or two points  
         (set! *SELECT_INIT* (cons 'no_empty *SELECT_INIT*))
         (set! *PICK_MODE* 'select_coord)
         (gui-entry-label-set-text *SELECT_TEXT_PREFIX*)
         ; stop the signal as well, we are in a callback function
         (gui-entry-callback-replace selection-dist-point)
         
         (gui-option-menu-set *OPTION_POINT*)
    )
    ((dist)
         ; write back the displacement vector
         (gui-entry-add-history 
           (string-append "\nDistance vector: "
                          (number->string *SELECT_X*) " "
                          (number->string *SELECT_Y*) " "
                          (number->string *SELECT_Z*)))

         (undo-add-op-mark)
         ; copy the objects
         (object-copy-internal)

         ; clear all selection variables
         (selection-clear)
         ; update the screen
         (gui-update-gl-window #t)
    )
    ((cancel quit)
         ; clear all selection variables
         (selection-clear)
         ; update the screen
         ; removes selection coloring
         (gui-update-gl-window #t)
    )
  )
)



(define (operation-edit-property)
  (if (not (null? *SELECT_LIST*))
    (let*
      ; let the user decide what he wants
      ( (new-prop (gui-property-main-window 
                    (object-property-collect *SELECT_LIST*) #f)) )
      ; if the user accepted the edited properties
      (if new-prop
        (begin
          ; place the undo mark
          (undo-add-op-mark)
          (object-property-undo-update *SELECT_LIST* new-prop #t #f)
        )
      )
    )
  )

  ; clear all selection
  (selection-clear)
  ; update the screen
  (gui-update-gl-window #t)
)


(define (operation-getcoord)
  (if (equal? *SELECT_STATE* 'point)
    (begin
      (gui-entry-add-history 
        (string-append "\nThe coordinates of the specified point are: "
                       (number->string *SELECT_X*) " "
                       (number->string *SELECT_Y*) " "
                       (number->string *SELECT_Z*) )
      )
    )
  )

  ; clear all selection
  (selection-clear)
  ; update the screen
  (gui-update-gl-window #t)
)


(define (operation-getdist)
  (if (= (length *SELECT_LIST*) 2)
    (let*
      ( (elem1 (car *SELECT_LIST*))
        (elem2 (cadr *SELECT_LIST*))
        (coord1 (object-property-get (car elem1) (cadr elem1) 'coords))
        (coord2 (object-property-get (car elem2) (cadr elem2) 'coords))
        (dx (- (car coord1) (car coord2)))
        (dy (- (cadr coord1) (cadr coord2)))
        (dz (- (caddr coord1) (caddr coord2)))
        (dist (sqrt (+ (* dx dx) (* dy dy) (* dz dz))))
      )
      (gui-entry-add-history 
        (string-append "\nThe distance between the two points is: "
                       (number->string dist)
        )
      )
      ; clear all selection
      (selection-clear)
      ; update the screen
      (gui-update-gl-window #t)
    )
    (gui-entry-add-history " *Two points are required*")
  )
)


(define (operation-display-property)
  (case *SELECT_STATE*
    ; in case of ENTER clear selection and return
    ( (cancel)
      (selection-clear)
      (gui-update-gl-window #t)
    )
    ; the name of the property is supplied
    ( (string)
      (let* 
        ( (prop (string->symbol *SELECT_ANSWER*))
          (sel-list '())
        )
        ; go through all element types
        (for-each
          (lambda (obj-type obj-list)
            ; go through all elements
            (for-each
              (lambda (elem)
                ; if the property is present add it to the list
                (if (assoc prop (cdr elem))
                  (set! sel-list (cons (list obj-type (car elem)) sel-list))
                )
              )
              obj-list
            )
          )
          '(p l t q b g)
          (list *POINTS* *LINKS* *TRIANGS* *QUADS* *BOXES* *GROUPS*)
        )
        ; write back how many elements are found
        (gui-entry-add-history 
          (string-append "\nNumber of objects: " 
                         (number->string (length sel-list))
          )
        )
        ; make this the current selection
        (set! *SELECT_LIST* sel-list)
        ; update the screen
        (gui-update-gl-window #t)
      )
    )
  )
)

