(define-module (sx selection-basic))
(use-modules (sx gui-entry) (sx gui-optmenu))
(export *SELECT_PREVIOUS* *SELECT_ACTION*       
        *SELECT_USER_TYPE* *SELECT_USER_SINGLE* 
        *SELECT_STACK*
;;;;;;;;
        selection-push-internal selection-pop-internal
        selection-keyword-internal
;;;;;;;;
        selection-crossing-enter selection-crossing-quit
        selection-setup
        selection-clear
        selection-return
        selection-keyword selection-integer selection-real selection-string)

;----------------------------------------------------------------------------

; FIXME: should be renamed
; all dialog windows must register here
; this will stop the ESC reset and menu reset to interfere
(define-public *DIALOG_WINDOWS* '())


; FIXME: should be renamed
; It determines the behaviour of the mouse.
; Possible values:
; view, select_coord, select_obj, select_cross
(define-public *PICK_MODE* 'view)

; the list of keywords
(define-public *SELECT_KEYWORDS* '())
; can contain 
; - no_negative no_zero no_empty no_multiple
; - type_any 
; - type_point type_link type_triang type_quad type_box
; - type_group
(define-public *SELECT_INIT* '())

; the state of the selection
; cancel, quit, 
; objects, point, integer, real, keyword, dist
(define-public *SELECT_STATE* 'init)
; mode of selection: add, remove, invert
(define-public *SELECT_OBJ_MODE* 'invert)
; by default the prefix is for point selection
(define-public *SELECT_PREFIX* 'p)
; text prefix, to add extra text if necessary to guide the selection
(define-public *SELECT_TEXT_PREFIX* #f)



; stores the first corner of a selection box
(define-public *SELECT_WIN_CORNER_1* #f)
; the other corner of the selection box
; (This may change as the mouse is moving around the screen.)
(define-public *SELECT_WIN_CORNER_2* #f)


; -------------------------------------------
; Selection results
;--------------------------------------------
; storing the selected points coordinates
(define-public *SELECT_X* #f)
(define-public *SELECT_Y* #f)
(define-public *SELECT_Z* #f)
; base point for distance
(define-public *SELECT_BASE* #f)
; the result of the get functions (keyword, integer, real)
(define-public *SELECT_ANSWER* #f)
; selected object list
(define-public *SELECT_LIST* '())


; -----------------------------------------------------------------------------
; previous accepted selection of objects by the user
(define *SELECT_PREVIOUS* '())

; action to take with selection
(define *SELECT_ACTION* #f)

; user can restrict the selectable element types
; by default, any element can be selected
; Possible values:
;   any, p, l, ...
; It is restrictive, that is why not a list like SELECT_INIT
(define *SELECT_USER_TYPE* 'any)

; user can restrict that he/she wants to select only one element
; by default multiple selection
(define *SELECT_USER_SINGLE* #f)

; when a selection mechanism calls another selection, 
; it can push its state variables on this stack
(define *SELECT_STACK* '())

; -----------------------------------------------------------------------------

(define (selection-init-internal)
  (set! *SELECT_LIST* '())         ; no selected object
  (set! *SELECT_X* #f)             ; no coordinates are selected
  (set! *SELECT_Y* #f)
  (set! *SELECT_Z* #f)
  (set! *SELECT_BASE* #f)          ; no base point for distance
  (set! *SELECT_ANSWER* #f)        ; initial answer is nothing

  (set! *SELECT_STATE* 'init)      ; initial state
  (set! *SELECT_PREFIX* 'p)        ; initial selection prefix is point
  (set! *SELECT_OBJ_MODE* 'invert) ; inverting selection mode

  (set! *SELECT_KEYWORDS* '())     ; no keywords 
  (set! *SELECT_INIT* '())         ; no controlling objects

  ; no selection by window
  (set! *SELECT_WIN_CORNER_1* #f)
  (set! *SELECT_WIN_CORNER_2* #f)

  ; user restrictions
  (set! *SELECT_USER_TYPE* 'any)
  (set! *SELECT_USER_SINGLE* #f)
)


(define (selection-clear-internal)
  ; clear variables
  (selection-init-internal)

  (gui-entry-callback-replace gui-entry-default-callback)
  (set! *SELECT_ACTION* #f)        ; no action to take
  (set! *SELECT_TEXT_PREFIX* #f)   ; no text during selection

  ; no optional menu
  (gui-option-menu-set *OPTION_EMPTY*)

  ; reset the graphics mode to simple viewing
  (set! *PICK_MODE* 'view)

  ; reset the entry text to default
  (gui-entry-label-set-text *ENTRY_DEFAULT_TEXT*)
)

(define (selection-push-internal)
  (set! *SELECT_STACK* 
    (cons (list *SELECT_LIST*
                *SELECT_X* *SELECT_Y* *SELECT_Z*
                *SELECT_BASE* *SELECT_ANSWER*
                *SELECT_STATE* *SELECT_PREFIX* 
                *SELECT_KEYWORDS* *SELECT_INIT*
                *SELECT_WIN_CORNER_1* *SELECT_WIN_CORNER_2*
                *SELECT_USER_TYPE* *SELECT_USER_SINGLE*
                ;
                *SELECT_TEXT_PREFIX*
                *SELECT_OBJ_MODE*
                *SELECT_ACTION*
                *PICK_MODE*
                ; actual printed text
                (gui-entry-label-get-text)
                ; current callback
                *ENTRY_CALLBACK*
                ; current optional menu
                (gui-option-menu-get)
          )
          *SELECT_STACK*
    )
  )
  ; reset the "upper" variables
  (selection-init-internal)
)

(define (selection-pop-internal)
  (let* 
    ( (sellist (car *SELECT_STACK*)) )
    (set! *SELECT_STACK* (cdr *SELECT_STACK*))

    (set! *SELECT_LIST* (list-ref sellist 0))
    (set! *SELECT_X*    (list-ref sellist 1))
    (set! *SELECT_Y*    (list-ref sellist 2))
    (set! *SELECT_Z*    (list-ref sellist 3))

    (set! *SELECT_BASE*   (list-ref sellist 4))
    (set! *SELECT_ANSWER* (list-ref sellist 5))

    (set! *SELECT_STATE*  (list-ref sellist 6))
    (set! *SELECT_PREFIX* (list-ref sellist 7))

    (set! *SELECT_KEYWORDS* (list-ref sellist 8))
    (set! *SELECT_INIT*     (list-ref sellist 9))

    (set! *SELECT_WIN_CORNER_1* (list-ref sellist 10))
    (set! *SELECT_WIN_CORNER_2* (list-ref sellist 11))

    (set! *SELECT_USER_TYPE*    (list-ref sellist 12))
    (set! *SELECT_USER_SINGLE*  (list-ref sellist 13))

    (set! *SELECT_TEXT_PREFIX*  (list-ref sellist 14))
    (set! *SELECT_OBJ_MODE*     (list-ref sellist 15))
    (set! *SELECT_ACTION*       (list-ref sellist 16))
    (set! *PICK_MODE*           (list-ref sellist 17))

    (gui-entry-label-set-text   (list-ref sellist 18))
    (gui-entry-callback-replace (list-ref sellist 19))
    (gui-option-menu-set        (list-ref sellist 20))
  )
)


(define (selection-setup-internal value keywords)
  (if (list? value)
    (set! *SELECT_INIT* value)
    (set! *SELECT_INIT* '())
  )

  (set! *SELECT_KEYWORDS* '())
  (if keywords
    (let* 
      ( (words (text-split-with-empty keywords #\space #f))
        (pair #f) (end 0) (keylength #f)
      )
      (for-each
        (lambda (x)
          ; try to split into keyword - shortcut pair
          (set! pair (text-split-with-empty x #\, #f))
          ; keyword cannot be empty
          (if (string=? (car pair) "")
            (begin
              (gui-entry-add-history 
                "\n*Empty keyword found in initialisation*")
            )
            ; check that keyword - shortcut pair was given
            (if (and (= (length pair) 2)
                     (not (string=? (cadr pair) "")) )
              (begin
                ; check that shortcut matches the keyword
                (if (string-ci=? (cadr pair) 
                                 (substring (car pair) 
                                            0 
                                            (string-length (cadr pair))))
                  (set! *SELECT_KEYWORDS* (cons pair *SELECT_KEYWORDS*))
                  (begin
                    (gui-entry-add-history 
                      "\n*Keyword and shortcut do not match*")
                  )
                )
              )
              ; the beginning of keyword must be capitalised
              (begin
                (set! end 0)
                (do ((i 0 (1+ i))) 
                    ((or (= i (string-length (car pair)))
                         (char-lower-case? (string-ref (car pair) i))))
                  (set! end (1+ end))
                )
                (if (> end 0)
                  (set! *SELECT_KEYWORDS* 
                    (cons (list (car pair) (substring (car pair) 0 end))
                          *SELECT_KEYWORDS*))
                  (set! *SELECT_KEYWORDS* 
                    (cons (list (car pair) (car pair))
                          *SELECT_KEYWORDS*))
                )
              )
            )
          )
        )
        words
      )
    )
  )
)



;;; ---------------------------------------------------------------------------


; returns the value in *SELECT_ANSWER*
(define (selection-keyword-internal selection action)
  (if (= (string-length selection) 0)
    (begin
      (if (member 'no_empty *SELECT_INIT*)
        (begin
          (gui-entry-add-history " *Keyword is required*")
          (set! *SELECT_STATE* 'init)
        )
        (begin
          (set! *SELECT_STATE* 'cancel)
          (set! *SELECT_ANSWER* #f)
          (action)
        )
      )
    )
    (let* ((answer #f) (amblist '()))
      ; try to find as a shortcut
      (for-each
        (lambda (x)
          (if (string-ci=? selection (cadr x))
            (set! answer (car x))
          )
        )
        *SELECT_KEYWORDS*
      )
      (if (not answer)
        (begin
          ; collect matching keywords
          (for-each (lambda (x)
            (if (and (<= (string-length selection) (string-length (car x)))
                     (string-ci=? selection 
                             (substring (car x) 0 (string-length selection))) )
              (set! amblist (cons (car x) amblist))
            ) )
            *SELECT_KEYWORDS*
          )
          ; determine what to do
          (case (length amblist)
            ( (0)
              (gui-entry-add-history " *Invalid keyword*")
              (set! *SELECT_STATE* 'init) )
            ( (1)
              (set! answer (car amblist)) )
            ( else
              (gui-entry-add-history " Ambiguous answer, please clarify...")
              (for-each (lambda (x)
                (gui-entry-add-history (string-append " " x)) )
                amblist
              )
              (set! *SELECT_STATE* 'init)
            )
          )
        )
      )
      (if answer
        (begin
          ; write back the selected keyword
          (gui-entry-add-history (string-append " " answer))
          (set! *SELECT_STATE* 'keyword)
          (set! *SELECT_ANSWER* answer)
          (action)
        )
      )
    )
  )
)

; returns the value in *SELECT_ANSWER*
(define (selection-integer-internal selection action)
  (cond 
    ( (= (length selection) 0)
      (if (member 'no_empty *SELECT_INIT*)
        (begin
          (gui-entry-add-history " *Integer value is required*")
          (set! *SELECT_STATE* 'init)
        )
        (begin
          (set! *SELECT_STATE* 'cancel)
          (set! *SELECT_ANSWER* #f)
          (action)
        )
      )
    )
    ( (= (length selection) 1)
      ; check that it is an integer
      (if (integer? (car selection))
        (begin
          (cond
            ( (and (member 'no_negative *SELECT_INIT*)
                   (< (car selection) 0) )
              (gui-entry-add-history " *Positive integer value is required*")
              (set! *SELECT_STATE* 'init) )
            ( (and (member 'no_zero *SELECT_INIT*)
                   (= (car selection) 0) )
              (gui-entry-add-history " *Non-zero integer value is required*")
              (set! *SELECT_STATE* 'init) )
            ( else
              ; write back the integer
              (gui-entry-add-history (string-append " " (number->string (car selection))))
              (set! *SELECT_STATE* 'integer)
              (set! *SELECT_ANSWER* (car selection))
              (action) )
          )
        )
        (begin ; not integer
          (if (null? *SELECT_KEYWORDS*)
            (begin
              (gui-entry-add-history " *Invalid entry*") ; no keywords are possible
              (set! *SELECT_STATE* 'init)
            )
            (selection-keyword-internal (car selection) action)
          )
        )
      )
    )
    ( else (gui-entry-add-history " *Invalid entry*") 
           (set! *SELECT_STATE* 'init)
    )
  )
)

; returns the value in *SELECT_ANSWER*
(define (selection-real-internal selection action)
  (cond
    ( (= (length selection) 0)
      (if (member 'no_empty *SELECT_INIT*)
        (begin
          (gui-entry-add-history " *Real value is required*")
          (set! *SELECT_STATE* 'init)
        )
        (begin
          (set! *SELECT_STATE* 'cancel)
          (set! *SELECT_ANSWER* #f)
          (action)
        )
      )
    )
    ( (= (length selection) 1)
      ; check that it is a real number
      (if (real? (car selection))
        (begin
          (cond
            ( (and (member 'no_negative *SELECT_INIT*)
                   (< (car selection) 0) )
              (gui-entry-add-history " *Positive real value is required*")
              (set! *SELECT_STATE* 'init) )
            ( (and (member 'no_zero *SELECT_INIT*)
                   (= (car selection) 0) )
              (gui-entry-add-history " *Non-zero real value is required*")
              (set! *SELECT_STATE* 'init) )
            ( else
              ; write back the real number
              (gui-entry-add-history 
                (string-append " " (number->string (car selection))))
              (set! *SELECT_STATE* 'real)
              (set! *SELECT_ANSWER* (car selection))
              (action) )
          )
        )
        (begin ; not real
          (if (null? *SELECT_KEYWORDS*)
            (begin
              (gui-entry-add-history " *Invalid entry*") ; no keywords are possible
              (set! *SELECT_STATE* 'init)
            )
            (selection-keyword-internal (car selection) action)
          )
        )
      )
    )
    ( else (gui-entry-add-history " *Invalid entry*") 
           (set! *SELECT_STATE* 'init)
    )
  )
)




(define (selection-string-internal selection action)
  (if (= (string-length selection) 0)
    (begin
      (if (member 'no_empty *SELECT_INIT*)
        (begin
          (gui-entry-add-history " *String is required*")
          (set! *SELECT_STATE* 'init)
        )
        (begin
          (set! *SELECT_STATE* 'cancel)
          (set! *SELECT_ANSWER* #f)
          (action)
        )
      )
    )
    (begin
      ; print the user entered data
      (gui-entry-add-history (string-append " " selection))
      (set! *SELECT_STATE* 'string)
      (set! *SELECT_ANSWER* selection)
      (action)
    )
  )
)

;;; ---------------------------------------------------------------------------


(define (selection-crossing-enter text pt)
  (gui-entry-label-set-text text)
  (set! *PICK_MODE* 'select_cross)
  (set! *SELECT_WIN_CORNER_1* pt)
  (set! *SELECT_WIN_CORNER_2* #f)
)

; returns from crossing object selection to normal object selection
(define (selection-crossing-quit)
  (set! *SELECT_WIN_CORNER_1* #f)
  (set! *SELECT_WIN_CORNER_2* #f)
  (set! *PICK_MODE* 'select_obj)
  (gui-entry-label-set-text *SELECT_TEXT_PREFIX*)
)


(define (selection-setup callback action text opt_menu gl_mode
                         control keywords)
  ; clear variables
  (selection-init-internal)

  (gui-entry-callback-replace callback)
  (set! *SELECT_ACTION* action)    ; the action to take
  (set! *SELECT_TEXT_PREFIX* text) ; text during selection
  (gui-entry-label-set-text *SELECT_TEXT_PREFIX*)

  ; switch on an optional menu
  (if opt_menu
    (gui-option-menu-set opt_menu)
    (gui-option-menu-set *OPTION_EMPTY*)
  )

  ; selection style
  (set! *PICK_MODE* gl_mode)

  ; set control parameters and keywords for selection
  (selection-setup-internal control keywords)
)


; clears all variables and makes sure that no other waiting
; selection mechanism is on the stack
(define (selection-clear)
  ; if objects were selected and the selection is accepted (not canceled), 
  ; store it for later use
  (if (and (not (null? *SELECT_LIST*)) (not (eq? *SELECT_STATE* 'cancel)))
    (set! *SELECT_PREVIOUS* *SELECT_LIST*)
  )

  (selection-clear-internal)
  (set! *SELECT_STACK* '())
)

(define (selection-return)
  ; if objects were selected and the selection is accepted (not canceled), 
  ; store it for later use
  (if (and (not (null? *SELECT_LIST*)) (not (eq? *SELECT_STATE* 'cancel)))
    (set! *SELECT_PREVIOUS* *SELECT_LIST*)
  )

  (if (null? *SELECT_STACK*)
    ; no pushed selection, simply clear all variables
    (selection-clear-internal)
    ; if there is another selection mechanism on the stack 
    ; then restore its state and return to it
    (selection-pop-internal)
  )
)


(define (selection-keyword)
  (let*
    ( (tt (gui-entry-get-text)) )
    ; print out what the user entered
    (gui-entry-add-history (string-append "\n" (gui-entry-label-get-text) " "))

    (selection-keyword-internal tt *SELECT_ACTION*)
  )
)


(define (selection-integer)
  (let*
    ( (tt (gui-entry-get-text))
      (pp (map (lambda (x)
                 (if (integer? (string->number x)) (string->number x) x))
               (text-split-with-empty tt #\, #f)
          )
      )
    )
    ; print out the label
    (gui-entry-add-history (string-append "\n" (gui-entry-label-get-text) " "))
    ; do the job
    (selection-integer-internal pp *SELECT_ACTION*)
  )
)


(define (selection-real)
  (let*
    ( (tt (gui-entry-get-text))
      (pp (map (lambda (x)
                 (if (string->number x) (string->number x) x))
               (text-split-with-empty tt #\, #f)
          )
      )
    )
    ; print out the label
    (gui-entry-add-history (string-append "\n" (gui-entry-label-get-text) " "))
    ; do the job
    (selection-real-internal pp *SELECT_ACTION*)
  )
)


(define (selection-string)
  (let*
    ( (tt (gui-entry-get-text)) )
    ; print the question
    (gui-entry-add-history (string-append "\n" (gui-entry-label-get-text) " "))

    (selection-string-internal tt *SELECT_ACTION*)
  )
)

