(define-module (sx undo))

(export *CHANGED*
        redo-get undo-get-list undo-get-size undo-redo-reset
        redo-add undo-add
        undo-add-op-mark undo-add-user-mark
        undo-one-operation undo-all-operations undo-until-user-mark
        redo-operation)

(define *UNDO* '())    ; list of commands to undo
(define *REDO* '())    ; list of commands to redo
(define *CHANGED* #f)  ; marking that anything has changed

(define (redo-get)
  *REDO*
)

(define (undo-get-list)
  *UNDO*
)

(define (undo-get-size)
  (length *UNDO*)
)

(define (undo-redo-reset)
  (set! *REDO* '())
  (set! *UNDO* '())
)

;;; adding 

(define (redo-add com)
  (set! *REDO* (cons com *REDO*))
)

(define (undo-add com)
  (set! *UNDO* (cons com *UNDO*))
)

;;; marks

(define (undo-add-op-mark)
  (set! *CHANGED* #t)
  (undo-add 'OPERATION-MARK)
)

(define (undo-add-user-mark)
  (undo-add 'USER-MARK)
)

;;; operations on the undo, redo lists

(define (undo-one-operation)
  (set! *CHANGED* #t)
  ;; reset REDO, so only one operation can be redone
  (set! *REDO* '())
  (let* ((elem #f))
    ; go one back until it is empty or
    ; user mark or operation ends
    (do () ((or (null? *UNDO*)
                (eq? (car *UNDO*) 'OPERATION-MARK) ))
      ; just in case it encounters a user mark
      (if (not (eq? (car *UNDO*) 'USER-MARK))
        (primitive-eval (car *UNDO*))
      )
      (set! *UNDO* (cdr *UNDO*))
    )

    ; remove any mark sign, 
    ; if not at the end of the list
    (if (not (null? *UNDO*))
      (set! *UNDO* (cdr *UNDO*))
    )
  )
)


(define (undo-all-operations)
  (do () ((null? *UNDO*))
    (undo-one-operation)
  )
)

(define (undo-until-user-mark)
  ; everything until USER-MARK or empty list
  (do () ((or (null? *UNDO*)
              (eq? (car *UNDO*) 'USER-MARK) ))
    (undo-one-operation)
  )
  ; remove the mark sign itself, 
  ; if not at the end of the list
  (if (not (eq? *UNDO* '()))
    (set! *UNDO* (cdr *UNDO*))
  )
)


(define (redo-operation)
  (set! *CHANGED* #t)
  ; put an operation mark if we really redo something
  (if (not (null? *REDO*))
    (begin
      (undo-add-op-mark)
      ; redo until it is empty
      (do () ((null? *REDO*))
        (if (or (eq? (car *REDO*) 'USER-MARK)
                (eq? (car *REDO*) 'OPERATION-MARK) )
          (begin
            ; this never happens, unless it is corrupted
            (display "REDO BUG")(newline)
          )
          (primitive-eval (car *REDO*))
        )
        (set! *REDO* (cdr *REDO*))
      )
    )
  )
)







