(define-module (sx view))
(use-modules (sx property) (sx selection-basic) (sx selection) (sx render)
             (sx gui-gl) (sx gui-optmenu) (sx gui-entry)
)
(export view-setup show-setup view-reset view-center view-center-points
        view-style view-zoom view-light)

; local variables, the optional menus
(define *OPTION_VIEW*
  (gui-option-menu-new 
    '(("center" #t) ("style" #t) 
      ("reset" #t)  ("zoom" #t) ("light" #t))
  )
)

(define *OPTION_CENTER*
  (gui-option-menu-new 
    '(("point" #t) ("object" #t) ("all" #t))
  )
)

(define *OPTION_RESET*
  (gui-option-menu-new 
    '(("all" #t) ("rotation" #t) ("zoom" #t) ("pan" #t))
  )
)

(define *OPTION_STYLE*
  (gui-option-menu-new
    '(("all" #t) ("point" #t) ("link" #t) ("triang" #t)
      ("quad" #t) ("box" #t) ("group" #t))
  )
)

(define *OPTION_APPEAR_ALL*
  (gui-option-menu-new
    '(("opaq" #t) ("transparent" #t) ("lines" #t)
      ("full" #t) ("tiled" #t) ("show" #t) ("hide" #t))
  )
)

(define *OPTION_APPEAR_POINT*
  (gui-option-menu-new
    '(("show" #t) ("hide" #t))
  )
)

(define *OPTION_APPEAR_LINK*
  (gui-option-menu-new
    '(("full" #t) ("tiled" #t) ("show" #t) ("hide" #t))
  )
)

(define *OPTION_APPEAR_BOX*
  (gui-option-menu-new
    '(("opaq" #t) ("transparent" #t) ("lines" #t)
      ("show" #t) ("hide" #t))
  )
)



(define (view-setup)
  (selection-setup selection-keyword
                   view-internal
                   "Press ENTER to finish viewing or [Center/Reset/Style/Light/Zoom]"
                   *OPTION_VIEW*
                   'view
                   #f 
                   "Center Reset Style Light Zoom")
)


(define (view-internal)
  (case *SELECT_STATE*
    ( (cancel)
      (selection-return)
      ; display list must be regenerated, 
      ; since there may be an object selection
;      (gui-update-gl-window #t)
    )
    ( (keyword)
      (cond
        ( (string-ci=? *SELECT_ANSWER* "Center")
          (selection-push-internal)
          (view-center)
          (gui-update-gl-window #t)
        )
        ( (string-ci=? *SELECT_ANSWER* "Reset")
          (selection-push-internal)
          (view-reset)
          (gui-update-gl-window #t)
        )
        ( (string-ci=? *SELECT_ANSWER* "Style")
          (selection-push-internal)
          (view-style)
          (gui-update-gl-window #t)
        )
        ( (string-ci=? *SELECT_ANSWER* "Light")
          (selection-push-internal)
          (view-light)
          (gui-update-gl-window #t)
        )
        ( (string-ci=? *SELECT_ANSWER* "Zoom")
          (view-zoom)
        )
      )
    )
  )
)


(define (show-setup)
  (selection-setup selection-keyword
                   show-internal
                   "Viewing or [Center/Reset/Style/Light/Zoom]"
                   *OPTION_VIEW*
                   'view
                   '(no_empty)
                   "Center Reset Style Light Zoom")
)



(define (show-internal)
  (case *SELECT_STATE*
    ( (keyword)
      (cond
        ( (string-ci=? *SELECT_ANSWER* "Center")
          (selection-push-internal)
          (gui-update-gl-window #t)
          (view-center)
        )
        ( (string-ci=? *SELECT_ANSWER* "Reset")
          (selection-push-internal)
          (gui-update-gl-window #t)
          (view-reset)
        )
        ( (string-ci=? *SELECT_ANSWER* "Style")
          (selection-push-internal)
          (gui-update-gl-window #t)
          (view-style)
        )
        ( (string-ci=? *SELECT_ANSWER* "Light")
          (selection-push-internal)
          (gui-update-gl-window #t)
          (view-light)
        )
        ( (string-ci=? *SELECT_ANSWER* "Zoom")
          (view-zoom)
        )
      )
    )
  )
)


(define (view-reset)
  (selection-setup selection-keyword
                   view-reset-internal
                   "Reset view [All/Rotation/Zoom/Pan]:"
                   *OPTION_RESET* 'view #f 
                   "All Rotation Zoom Pan")
)



(define (view-reset-internal)
  (case *SELECT_STATE*
    ( (keyword)
      (cond
        ( (string-ci=? *SELECT_ANSWER* "All")
          (set! *ZOOM* 1.0)
          (set! *QUATERNION* '(0 0 0 1.0))
          (set! *XDIST* 0.0)
          (set! *YDIST* 0.0)
        )
        ( (string-ci=? *SELECT_ANSWER* "Rotation")
          (set! *QUATERNION* '(0 0 0 1.0))
        )
        ( (string-ci=? *SELECT_ANSWER* "Zoom")
          (set! *ZOOM* 1.0)
        )
        ( (string-ci=? *SELECT_ANSWER* "Pan")
          (set! *XDIST* 0.0)
          (set! *YDIST* 0.0)
        )
      )
    )
  )
  ; clear keywords
  (selection-return)
  ; since it is reentrant, the screen must be properly updated
  (gui-update-gl-window #t)
)


(define (view-center)
  (selection-setup selection-keyword
                   view-center-internal
                   "New rotation center of [Point/Object center/All]:"
                   *OPTION_CENTER*
                   'view #f
                   "Point Object All")
)


(define (view-center-points points)
  (if (null? points)
    (set! *GL_CENTER* '(0.0 0.0 0.0))
    (let*
      ( (xc 0.0) (yc 0.0) (zc 0.0) (co #f) )
      (for-each (lambda (x)
          (set! co (object-property-get 'p (car x) 'coords))
          (set! xc (+ xc (car co)))
          (set! yc (+ yc (cadr co)))
          (set! zc (+ zc (caddr co)))
        )
        points
      )
      (set! *GL_CENTER* (list (/ xc (length points))
                              (/ yc (length points))
                              (/ zc (length points))))
    )
  )
  (set! *XDIST* 0.0)
  (set! *YDIST* 0.0)
)



(define (view-center-internal)
  (case *SELECT_STATE*
    ; when the point is supplied make it the new centre
    ( (point)
      (set! *GL_CENTER* (list *SELECT_X* *SELECT_Y* *SELECT_Z*))
      (set! *XDIST* 0.0)
      (set! *YDIST* 0.0)
      (selection-return)
      (gui-update-gl-window #t)
    )
    ( (objects)
      (let* ( (points '()) )
        ; add all points of selected objects
        (selection-add-ref-nodes #t)
        (for-each 
          (lambda (x) 
            (if (equal? (car x) 'p) 
              (set! points (cons (cdr x) points))
            )
          )
          *SELECT_LIST*
        )
        ; calculate the centre of the point "cloud"
        (view-center-points points)
      )

      (selection-return)
      (gui-update-gl-window #t)
    )
    ; This is the first stage
    ( (keyword)
      (cond 
        ; centre of all of the objects
        ( (string-ci=? *SELECT_ANSWER* "All")
          (view-center-points *POINTS*)
          (selection-return)
          (gui-update-gl-window #t)
        )
        ; centre point is explicitly specified
        ( (string-ci=? *SELECT_ANSWER* "Point")
          ; change selection mode
          ; FIXME: ugly, should be with a function
          (set! *SELECT_STATE* 'init)
          (set! *SELECT_TEXT_PREFIX* "Point of rotation centre:")
          (set! *SELECT_INIT* (list 'no_empty))
          (set! *PICK_MODE* 'select_coord)
          (gui-entry-label-set-text *SELECT_TEXT_PREFIX*)
          (gui-entry-callback-replace selection-coord)
          (gui-option-menu-set *OPTION_POINT*)
        )
        ; centre of the selected objects
        ( (string-ci=? *SELECT_ANSWER* "Object")
          ; change selection mode
          ; FIXME: ugly, should be with a function
          (set! *SELECT_STATE* 'init)
          (set! *SELECT_TEXT_PREFIX* "Select object(s):")
          (set! *SELECT_INIT* (list 'no_empty 'type_any))
          (set! *PICK_MODE* 'select_obj)
          (gui-entry-label-set-text *SELECT_TEXT_PREFIX*)
          (gui-entry-callback-replace selection-object)
          (gui-option-menu-set *OPTION_SELECT*)
        )
      )
    )
    ; default action, nothing
    ( (cancel)
      (selection-return)
      (gui-update-gl-window #t)
    )
  )

)

(define (view-style)
  (selection-setup selection-keyword
                   render-style-internal
                   (string-append 
                      "Rendering style for "
                      "[All/Point/Link/Triang/Quad/Box/Group] <All>")
                   *OPTION_STYLE* 'view #f
                   "All Point Link Triang Quad Box Group")
)


(define operation-render-object #f)

(define (render-style-internal)
  (case *SELECT_STATE*
    ( (cancel)
      ; for all of the object types
      (selection-setup selection-real
                       operation-render-style-all
                       (string-append
                         "Tiling percent [0-1.0] or "
                         "[Opaq/TRansparent/Lines/Full/TIled/Show/Hide] " 
                       )
                       *OPTION_APPEAR_ALL* 'view '(no_negative)
                       "Opaq TRansparent Lines Full TIled Show Hide"
      )
    )
    ( (keyword)
      (cond
        ( (string-ci=? *SELECT_ANSWER* "All")
          (selection-setup selection-real
                           operation-render-style-all
                           (string-append
                             "Tiling percent [0-1.0] or "
                             "[Opaq/TRansparent/Lines/Full/TIled/Show/Hide] " 
                           )
                           *OPTION_APPEAR_ALL* 'view '(no_negative)
                           "Opaq TRansparent Lines Full TIled Show Hide"
          )
        )
        ( (string-ci=? *SELECT_ANSWER* "Point")
          (set! operation-render-object 'p)
          (selection-setup selection-keyword
                           operation-render-style-per-object
                           "Select [Show or Hide]"
                           *OPTION_APPEAR_POINT* 'view #f
                           "Show Hide")
        )
        ( (string-ci=? *SELECT_ANSWER* "Link")
          (set! operation-render-object 'l)
          (selection-setup selection-real
                           operation-render-style-per-object
                           (string-append
                             "Tiling percent or "
                             "[Full/TIled/Show/Hide] <" 
                             (number->string 
                               (render-style-get 'l 'tile_percent))
                             ">" )
                           *OPTION_APPEAR_LINK* 'view '(no_negative)
                           "Full TIled Show Hide")
        )
        ( (string-ci=? *SELECT_ANSWER* "Triang")
          (set! operation-render-object 't)
          (selection-setup selection-real
                           operation-render-style-per-object
                           (string-append
                             "Tiling percent or "
                             "[Opaq/TRansparent/Lines/Full/TIled/Show/Hide] <" 
                             (number->string 
                               (render-style-get 't 'tile_percent))
                             ">" )
                           *OPTION_APPEAR_ALL* 'view '(no_negative)
                           "Opaq TRansparent Lines Full TIled Show Hide")
        )
        ( (string-ci=? *SELECT_ANSWER* "Quad")
          (set! operation-render-object 'q)
          (selection-setup selection-real
                           operation-render-style-per-object
                           (string-append
                             "Tiling percent or "
                             "[Opaq/TRansparent/Lines/Full/TIled/Show/Hide] <" 
                             (number->string (render-style-get 'q 'tile_percent))
                             ">" )
                           *OPTION_APPEAR_ALL* 'view '(no_negative)
                           "Opaq TRansparent Lines Full TIled Show Hide")
        )
        ( (string-ci=? *SELECT_ANSWER* "Box")
          (set! operation-render-object 'b)
          (selection-setup selection-keyword
                           operation-render-style-per-object
                           "Select [Opaq/Transparent/Lines/Show/Hide] "
                           *OPTION_APPEAR_BOX* 'view #f
                           "Opaq Transparent Lines Show Hide")
        )
        ( (string-ci=? *SELECT_ANSWER* "Group")
          (set! operation-render-object 'q)
          (selection-setup selection-real
                           operation-render-style-per-object
                           (string-append
                             "Tiling percent or "
                             "[Opaq/TRansparent/Lines/Full/TIled/Show/Hide] <" 
                             (number->string 
                               (render-style-get 'g 'tile_percent))
                             ">" )
                           *OPTION_APPEAR_ALL* 'view '(no_negative)
                           "Opaq TRansparent Lines Full TIled Show Hide")
        )
      )
    )
  )
)


(define (operation-render-style-per-object)
  (case *SELECT_STATE*
    ( (real)
      (if (and (>= *SELECT_ANSWER* 0.0) (<= *SELECT_ANSWER* 1.0))
        (render-style-set operation-render-object 'tile_percent *SELECT_ANSWER*)
        (gui-entry-add-history "\n*Invalid tiling percent [0.0 - 1.0]*")
      )
    )
    ( (keyword)
      (cond
        ( (string-ci=? *SELECT_ANSWER* "Opaq")
          (render-style-set operation-render-object 'style 'opaq)
        )
        ( (string-ci=? *SELECT_ANSWER* "Transparent")
          (render-style-set operation-render-object 'style 'transparent)
        )
        ( (string-ci=? *SELECT_ANSWER* "Lines")
          (render-style-set operation-render-object 'style 'lines)
        )
        ( (string-ci=? *SELECT_ANSWER* "Tiled")
          (render-style-set operation-render-object 'tiling #t)
        )
        ( (string-ci=? *SELECT_ANSWER* "Full")
          (render-style-set operation-render-object 'tiling #f)
        )
        ( (string-ci=? *SELECT_ANSWER* "Show")
          (render-style-set operation-render-object 'show #t)
        )
        ( (string-ci=? *SELECT_ANSWER* "Hide")
          (render-style-set operation-render-object 'show #f)
        )
      )
    )
  )
  ; clear keywords
  (selection-return)
  ; update the screen
  (gui-update-gl-window #t)
)


(define (operation-render-style-all)
  (case *SELECT_STATE*
    ( (real)
      (if (and (>= *SELECT_ANSWER* 0.0) (<= *SELECT_ANSWER* 1.0))
        (for-each 
          (lambda (type) (render-style-set type 'tile_percent *SELECT_ANSWER*) )
          '(l t q b g)
        )
        (gui-entry-add-history "\n*Invalid tiling percent [0.0 - 1.0]*")
      )
    )
    ( (keyword)
      (cond
        ( (string-ci=? *SELECT_ANSWER* "Opaq")
          (for-each 
            (lambda (type) (render-style-set type 'style 'opaq) )
            '(t q b g)
          )
        )
        ( (string-ci=? *SELECT_ANSWER* "Transparent")
          (for-each 
            (lambda (type) (render-style-set type 'style 'transparent) )
            '(t q b g)
          )
        )
        ( (string-ci=? *SELECT_ANSWER* "Lines")
          (for-each 
            (lambda (type) (render-style-set type 'style 'lines) )
            '(t q b g)
          )
        )
        ( (string-ci=? *SELECT_ANSWER* "Tiled")
          (for-each 
            (lambda (type) (render-style-set type 'tiling #t) )
            '(l t q b g)
          )
        )
        ( (string-ci=? *SELECT_ANSWER* "Full")
          (for-each 
            (lambda (type) (render-style-set type 'tiling #f) )
            '(l t q b g)
          )
        )
        ( (string-ci=? *SELECT_ANSWER* "Show")
          (for-each 
            (lambda (type) (render-style-set type 'show #t) )
            '(p l t q b g)
          )
        )
        ( (string-ci=? *SELECT_ANSWER* "Hide")
          (for-each 
            (lambda (type) (render-style-set type 'show #f) )
            '(p l t q b g)
          )
        )
      )
    )
  )
  ; clear keywords
  (selection-return)
  ; update the screen
  (gui-update-gl-window #t)
)


(define (view-zoom)
  (set! *ZOOM* (/ 0.9 *POINT_MAX*))

  ; the screen must be updated to reflect the change
  (gui-update-gl-window #t)
)


(define (view-light)
  (selection-setup selection-keyword
                   light-internal
                   "Lights? [Yes/No] "
                   *OPTION_YES_NO* 'view #f
                   "Yes No")
)


(define (light-internal)
  (case *SELECT_STATE*
    ( (keyword)
      (cond
        ( (string-ci=? *SELECT_ANSWER* "Yes")
          (set! *GL_RENDER_LIGHTING* #t)
        )
        ( (string-ci=? *SELECT_ANSWER* "No")
          (set! *GL_RENDER_LIGHTING* #f)
        )
      )
    )
  )
  ; clear keywords
  (selection-return)
  ; update the screen
  (gui-update-gl-window #t)
)







